#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import re
import warnings

from obspy import UTCDateTime, read_events
from obspy.core.inventory import Inventory, Network, Station, Channel
from obspy.core.util import NamedTemporaryFile
from obspy.core.util.testing import compare_xml_strings, remove_unique_ids
from obspy.io.nlloc.core import is_nlloc_hyp, read_nlloc_hyp, write_nlloc_obs


def _mock_coordinate_converter(x, y, z):
    """
    Mocks the following pyproj based converter function for the values
    encountered in the test. Mocks the following function::

        import pyproj
        proj_wgs84 = pyproj.Proj("epsg:4326")
        proj_gk4 = pyproj.Proj("epsg:31468")
        def my_conversion(x, y, z):
            x, y = pyproj.transform(proj_gk4, proj_wgs84, x * 1e3, y * 1e3)
            return x, y, z
    """
    if (x, y, z) == (4473.68, 5323.28, 4.57949):
        return (11.6455375456446, 48.04707051747388, 4.57949)
    else:
        raise Exception("Unexpected values during test run.")


class TestNLLOC():
    """
    Test suite for obspy.io.nlloc
    """
    def test_write_nlloc_obs(self, testdata):
        """
        Test writing nonlinloc observations phase file.
        """
        # load nlloc_custom.qml QuakeML file to generate OBS file from it
        filename = testdata["nlloc_custom.qml"]
        cat = read_events(filename, "QUAKEML")
        # adjust one pick time that got cropped by nonlinloc in NLLOC HYP file
        # due to less precision in hypocenter file (that we used to create the
        # reference QuakeML file)
        for pick in cat[0].picks:
            if pick.waveform_id.station_code == "UH4" and \
               pick.phase_hint == "P":
                pick.time -= 0.005

        # read expected OBS file output
        filename = testdata["nlloc.obs"]
        with open(filename, "rb") as fh:
            expected = fh.read().decode()

        # write via plugin
        with NamedTemporaryFile() as tf:
            cat.write(tf, format="NLLOC_OBS")
            tf.seek(0)
            got = tf.read().decode()

        assert expected == got

        # write manually
        with NamedTemporaryFile() as tf:
            write_nlloc_obs(cat, tf)
            tf.seek(0)
            got = tf.read().decode()

        assert expected == got

    def test_read_nlloc_hyp(self, testdata):
        """
        Test reading nonlinloc hypocenter phase file.
        """
        filename = testdata["nlloc_custom.hyp"]
        cat = read_nlloc_hyp(filename,
                             coordinate_converter=_mock_coordinate_converter)
        # reset pick channel codes, these got automatically mapped upon reading
        for pick in cat[0].picks:
            pick.waveform_id.channel_code = None
        with open(testdata["nlloc_custom.qml"], 'rb') as tf:
            quakeml_expected = tf.read().decode()
        with NamedTemporaryFile() as tf:
            cat.write(tf, format="QUAKEML")
            tf.seek(0)
            quakeml_got = tf.read().decode()

        # test creation times manually as they get omitted in the overall test
        creation_time = UTCDateTime("2014-10-17T16:30:08.000000Z")
        assert cat[0].creation_info.creation_time == creation_time
        assert cat[0].origins[0].creation_info.creation_time == creation_time

        quakeml_expected = remove_unique_ids(quakeml_expected,
                                             remove_creation_time=True)
        quakeml_got = remove_unique_ids(quakeml_got, remove_creation_time=True)
        # In python 3 float.__str__ outputs 5 decimals of precision more.
        # We use it in writing QuakeML, so files look different on Py2/3.
        # We use regex to cut off floats in the xml such that we only compare
        # 7 digits.
        pattern = r'(<.*?>[0-9]*?\.[0-9]{7})[0-9]*?(</.*?>)'
        quakeml_expected = re.sub(pattern, r'\1\2', quakeml_expected)
        quakeml_got = re.sub(pattern, r'\1\2', quakeml_got)

        # remove (changing) obspy version number from output
        re_pattern = '<version>ObsPy .*?</version>'
        quakeml_expected = re.sub(re_pattern, '', quakeml_expected, count=1)
        quakeml_got = re.sub(re_pattern, '', quakeml_got, count=1)

        compare_xml_strings(quakeml_expected.encode(), quakeml_got.encode())

    def test_read_nlloc_hyp_with_builtin_projection(self, testdata):
        """
        Test reading nonlinloc hyp file without a coordinate_converter.
        """
        cat = read_nlloc_hyp(testdata["nlloc.hyp"])
        cat_expected = read_events(testdata["nlloc.qml"])

        # test event
        ev = cat[0]
        ev_expected = cat_expected[0]
        assert round(abs(
            ev.creation_info.creation_time -
            ev_expected.creation_info.creation_time), 7) == 0

        # test origin
        orig = ev.origins[0]
        orig_expected = ev_expected.origins[0]
        assert round(abs(orig.time-orig_expected.time), 7) == 0
        assert round(abs(orig.longitude-orig_expected.longitude), 7) == 0
        assert round(abs(
            orig.longitude_errors.uncertainty -
            orig_expected.longitude_errors.uncertainty), 7) == 0
        assert round(abs(orig.latitude-orig_expected.latitude), 7) == 0
        assert round(abs(
            orig.latitude_errors.uncertainty -
            orig_expected.latitude_errors.uncertainty), 7) == 0
        assert round(abs(orig.depth-orig_expected.depth), 7) == 0
        assert round(abs(
            orig.depth_errors.uncertainty -
            orig_expected.depth_errors.uncertainty), 7) == 0
        assert round(abs(
            orig.depth_errors.confidence_level -
            orig_expected.depth_errors.confidence_level), 7) == 0
        assert orig.depth_type == orig_expected.depth_type
        assert orig.quality.associated_phase_count == \
            orig_expected.quality.associated_phase_count
        assert orig.quality.used_phase_count == \
            orig_expected.quality.used_phase_count
        assert orig.quality.associated_station_count == \
            orig_expected.quality.associated_station_count
        assert orig.quality.used_station_count == \
            orig_expected.quality.used_station_count
        assert round(abs(
            orig.quality.standard_error -
            orig_expected.quality.standard_error), 7) == 0
        assert round(abs(
            orig.quality.azimuthal_gap -
            orig_expected.quality.azimuthal_gap), 7) == 0
        assert round(abs(
            orig.quality.secondary_azimuthal_gap -
            orig_expected.quality.secondary_azimuthal_gap), 7) == 0
        assert orig.quality.ground_truth_level == \
            orig_expected.quality.ground_truth_level
        assert round(abs(
            orig.quality.minimum_distance -
            orig_expected.quality.minimum_distance), 7) == 0
        assert round(abs(
            orig.quality.maximum_distance -
            orig_expected.quality.maximum_distance), 7) == 0
        assert round(abs(
            orig.quality.median_distance -
            orig_expected.quality.median_distance), 7) == 0
        assert round(abs(
            orig.origin_uncertainty.min_horizontal_uncertainty -
            orig_expected.origin_uncertainty.min_horizontal_uncertainty), 7) \
            == 0
        assert round(abs(
            orig.origin_uncertainty.max_horizontal_uncertainty -
            orig_expected.origin_uncertainty.max_horizontal_uncertainty), 7) \
            == 0
        assert round(abs(
            orig.origin_uncertainty.azimuth_max_horizontal_uncertainty -
            getattr(orig_expected.origin_uncertainty,
                    "azimuth_max_horizontal_uncertainty")), 7) == 0
        assert orig.origin_uncertainty.preferred_description == \
            orig_expected.origin_uncertainty.preferred_description
        assert round(abs(
            orig.origin_uncertainty.confidence_level -
            orig_expected.origin_uncertainty.confidence_level), 7) == 0
        assert orig.creation_info.creation_time == \
            orig_expected.creation_info.creation_time
        assert orig.comments[0].text == orig_expected.comments[0].text

        # test a couple of arrivals
        for n in range(2):
            arriv = orig.arrivals[n]
            arriv_expected = orig_expected.arrivals[n]
            assert arriv.phase == arriv_expected.phase
            assert round(abs(arriv.azimuth-arriv_expected.azimuth), 7) == 0
            assert round(abs(arriv.distance-arriv_expected.distance), 7) == 0
            assert arriv.takeoff_angle is None
            assert arriv_expected.takeoff_angle is None
            assert round(
                abs(arriv.time_residual-arriv_expected.time_residual), 7) == 0
            assert round(
                abs(arriv.time_weight-arriv_expected.time_weight), 7) == 0

        # test a couple of picks
        for n in range(2):
            pick = ev.picks[n]
            pick_expected = ev_expected.picks[n]
            assert round(abs(pick.time-pick_expected.time), 7) == 0
            assert pick.waveform_id.station_code == \
                pick_expected.waveform_id.station_code
            assert pick.onset == pick_expected.onset
            assert pick.phase_hint == pick_expected.phase_hint
            assert pick.polarity == pick_expected.polarity

    def test_read_nlloc_hyp_via_plugin(self, testdata):
        filename = testdata["nlloc_custom.hyp"]
        cat = read_events(filename)
        assert len(cat) == 1
        cat = read_events(filename, format="NLLOC_HYP")
        assert len(cat) == 1

    def test_read_nlloc_with_pick_seed_id_lookup(self, testdata):
        # create some bogus metadata for lookup
        cha = Channel('HHZ', '00', 0, 0, 0, 0)
        sta = Station('HM02', 0, 0, 0, channels=[cha])
        cha = Channel('HHZ', '10', 0, 0, 0, 0)
        sta2 = Station('YYYY', 0, 0, 0, channels=[cha])
        net = Network('XX', stations=[sta, sta2])
        # second network with non matching data
        cha = Channel('HHZ', '00', 0, 0, 0, 0)
        sta = Station('ABCD', 0, 0, 0, channels=[cha])
        cha = Channel('HHZ', '10', 0, 0, 0, 0)
        sta2 = Station('EFGH', 0, 0, 0, channels=[cha])
        net2 = Network('YY', stations=[sta, sta2])

        inv = Inventory(networks=[net, net2], source='')

        filename = testdata["nlloc_custom.hyp"]
        # we get some warnings since we only provide sufficient metadata for
        # one pick
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            cat = read_events(filename, format="NLLOC_HYP", inventory=inv)
        assert len(cat) == 1
        for pick in cat[0].picks:
            wid = pick.waveform_id
            if wid.station_code == 'HM02':
                assert wid.network_code == 'XX'
                assert wid.location_code == ''
            else:
                assert wid.network_code == ''
                assert wid.location_code is None

    def test_is_nlloc_hyp(self, testdata):
        # test positive
        filename = testdata["nlloc_custom.hyp"]
        assert is_nlloc_hyp(filename)
        # test some negatives
        for filenames in ["nlloc_custom.qml", "nlloc.obs", "gaps.mseed",
                          "BW_RJOB.xml", "QFILE-TEST-ASC.ASC", "LMOW.BHE.SAC"]:
            filename = testdata["nlloc_custom.qml"]
            assert not is_nlloc_hyp(filename)

    def test_read_nlloc_with_picks(self, testdata):
        """
        Test correct resource ID linking when reading NLLOC_HYP file with
        providing original picks.
        """
        picks = read_events(testdata["nlloc_custom.qml"])[0].picks
        arrivals = read_events(
            testdata["nlloc_custom.hyp"], format="NLLOC_HYP",
            picks=picks)[0].origins[0].arrivals
        expected = [p.resource_id for p in picks]
        got = [a.pick_id for a in arrivals]
        assert expected == got

    def test_read_nlloc_with_multiple_events(self, testdata):
        """
        Test reading a NLLOC_HYP file with multiple hypocenters in it.
        """
        got = read_events(testdata["vanua.sum.grid0.loc.hyp"],
                          format="NLLOC_HYP")
        assert len(got) == 3
        assert got[0].origins[0].longitude == 167.049
        assert got[1].origins[0].longitude == 166.905
        assert got[2].origins[0].longitude == 166.858
        assert got[0].origins[0].latitude == -14.4937
        assert got[1].origins[0].latitude == -15.0823
        assert got[2].origins[0].latitude == -15.1529
        for item in got.events + [e.origins[0] for e in got.events]:
            assert item.creation_info.author == u'Océane Foix'
        for event in got.events:
            assert event.comments[0].text == \
                             "Central Vanuatu (3D tomo 2016)"

    def test_read_nlloc_6_beta_signature(self, testdata):
        """
        SIGNATURE field of nlloc hypocenter output file was somehow changed at
        some point after version 6.0 (it appears in 6.0.3 beta release for
        example).
        Date is now seemingly always prepended with 'run:' without a space
        afterwards.
        """
        filename = testdata['nlloc_post_version_6.hyp']
        cat = read_nlloc_hyp(filename)
        # check that signature time-of-run part is correctly read
        # (actually before the fix the above reading already fails..)
        assert cat[0].creation_info.creation_time == \
            UTCDateTime(2017, 5, 9, 11, 0, 22)

    def test_issue_2222(self, testdata):
        """
        Test that hour values of 24 don't break parser.
        """

        # modify the example file to contain an hour 24 and second 60
        with open(testdata['nlloc.hyp']) as f:
            nll_str = f.read().splitlines()
        # first add a line with hour 24
        str_list = list(nll_str[-3])
        str_list[37:41] = '2400'
        nll_str[-3] = ''.join(str_list)
        # then add a line with second 60
        str_list = list(nll_str[-4])
        str_list[46:48] = '60'
        nll_str[-4] = ''.join(str_list)
        # write to string io and read into catalog object
        str_io = io.StringIO()
        str_io.write('\n'.join(nll_str))
        str_io.seek(0)
        cat = read_nlloc_hyp(str_io)
        # check catalog is populated and pick times are right
        assert len(cat) == 1
        pick1, pick2 = cat[0].picks[-1], cat[0].picks[-2]
        assert pick1.time.hour == 0
        assert pick2.time.second == 0

    def test_reading_nlloc_v7_hyp_file(self, testdata):
        """
        Tests that we are getting the positioning of items in phase lines
        right. Values for arrivals are shifted by one index to the right in hyp
        files written by newer nonlinloc versions, see #3223
        Also tests that PUBLIC_ID line is correctly read and assigned to
        event.resource_id. Tests both literal "None" and actual IDs.
        """
        path = testdata['nlloc_v7.hyp']

        # Test 1: Read file with PUBLIC_ID None (literal string)
        # The literal "None" should not set resource_id,
        # so it gets auto-generated
        cat = read_nlloc_hyp(path)
        assert cat[0].origins[0].arrivals[0].azimuth == 107.42
        # compare test_rejected_origin test case
        assert cat[0].origins[0].evaluation_status is None
        ellipsoid = cat[0].origins[0].origin_uncertainty.confidence_ellipsoid
        assert ellipsoid is not None
        assert ellipsoid.semi_major_axis_length == 20.2574
        assert ellipsoid.semi_minor_axis_length == 9.1241
        assert ellipsoid.semi_intermediate_axis_length == 19.255
        assert ellipsoid.major_axis_plunge == 42.8141
        assert ellipsoid.major_axis_azimuth == 86.61
        assert ellipsoid.major_axis_rotation == 322.815
        # test that literal "None" leaves resource_id auto-generated
        auto_generated_id_1 = str(cat[0].resource_id)
        assert auto_generated_id_1.startswith('smi:local/')

        # Test 2: Dynamically modify PUBLIC_ID to a real value and test
        with open(path, 'rb') as fh:
            content = fh.read().decode()
        # Replace the PUBLIC_ID None line with a real ID
        modified_content = content.replace(
            'PUBLIC_ID None',
            'PUBLIC_ID smi:local/test-event-v7-public-id'
        )
        # Write to temporary file and read
        with io.BytesIO() as tmp:
            tmp.write(modified_content.encode())
            tmp.seek(0)
            cat2 = read_nlloc_hyp(tmp)
        # test that real PUBLIC_ID is read and set correctly
        assert str(cat2[0].resource_id) == "smi:local/test-event-v7-public-id"

    def test_rejected_origin(self, testdata):
        """
        Tests that we are marking rejected event/origin as such.
        Also tests that NLLOC header line is written into comment.
        (testing that evaluation status is left empty on "LOCATED" reported by
        nonlinloc is tested in other test case.
        """
        path = testdata['nlloc_rejected.hyp']
        cat = read_nlloc_hyp(path)
        expected_comment = (
            'NLLOC "./locs/20211214_2020-12-09_manual_loc/20211214_2020-12-09_'
            'manual.20201209.163708.grid0" "REJECTED" "WARNING: max prob '
            'location on grid boundary 10, rejecting location."')
        assert cat[0].origins[0].evaluation_status == "rejected"
        assert cat[0].origins[0].comments[1].text == expected_comment
        assert cat[0].comments[1].text == expected_comment

    def test_read_nlloc_doc_resolve_seedid(self):
        assert 'seedid_map:' in read_nlloc_hyp.__doc__
        assert 'ph2comp:' in read_nlloc_hyp.__doc__



# -*- coding: utf-8 -*-
"""
obspy.io.zmap - ZMAP read and write support for ObsPy
=====================================================

This module provides read and write support for the ZMAP format.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)


Usage Example
-------------

The ZMAP reader and writer hooks into the standard ObsPy event handling
mechanisms including format autodetection.

>>> from obspy.core.event import read_events
>>> cat = read_events('/path/to/zmap_events.txt')
>>> print(cat)
2 Event(s) in Catalog:
2012-04-04T14:21:42.300000Z | +41.818,  +79.689 | 4.4  None
2012-04-04T14:21:42.300000Z | +41.822,  +79.684 | 5.1  None
>>> cat.write('example.txt', format='ZMAP')  # doctest: +SKIP


Reading ZMAP
------------

Reading ZMAP is lenient, i.e. ``obspy.io.zmap`` will try to import a file
even if it doesn't strictly conform to 10 or 13 column ZMAP. Namely the
following deviations from standard ZMAP format are acceptable:

* Less or more than 10 or 13 columns. Extra columns are ignored. Missing
  values are set to ``None``.
* Integer years without a fractional part. If the fractional part is
  present, the date/time is computed from the year column. All other
  date/time fields are ignored.
  If the year column is an integer number, date and time are computed from all
  date/time related fields.

Note that ZMAP format autodetection in :func:`~obspy.core.event.read_events`
only works with strictly 10 or 13 column files. To read non-standard ZMAP
files, the ``format='ZMAP'`` keyword argument must be provided.

When reading ZMAP, the following mappings are used for uncertainties

* *Column 11 (Horizontal error):* stored in
  ``origin.origin_uncertainty.horizontal_uncertainty`` with
  ``preferred_description`` set accordingly.
* *Column 12 (Depth error):* stored in ``origin.depth_errors.uncertainty``
* *Column 13 (Magnitude error):* stored in
  ``magnitude.mag_errors.uncertainty``


Writing ZMAP
------------

When writing to ZMAP, the preferred origin and magnitude are used to fill the
origin and magnitude columns. Any missing values are exported as ``'NaN'``.

.. rubric:: Extended ZMAP

Writing extended ZMAP, is supported by using the keyword argument
``with_uncertainties``.

If :class:`~obspy.core.event.origin.OriginUncertainty` specifies a
*horizontal uncertainty* the value for column 11 is extracted from there.
*Uncertainty ellipse* and *confidence ellipsoid* are not currently supported.
If no horizontal uncertainty is given,
:class:`~obspy.core.event.origin.Origin`'s
``latitude_errors`` and ``longitude_errors`` are used instead. Depth and
magnitude errors are always read from the respective ``_errors`` attribute in
:class:`~obspy.core.event.origin.Origin`.


The ZMAP Format
---------------

ZMAP is a simple 10 column CSV file (technically TSV) format for basic catalog
data. It originates from ZMAP, a Matlab® based earthquake statistics package
(see [Wiemer2001]_). Since ZMAP files are purely numerical they are easily
imported into Matlab® using the ``dlmread`` function.

=================   ==============================================
Column #            Value
=================   ==============================================
 1                  Longitude [deg]
 2                  Latitude [deg]
 3                  Decimal year (e.g., 2005.5 for July 2nd, 2005)
 4                  Month
 5                  Day
 6                  Magnitude
 7                  Depth [km]
 8                  Hour
 9                  Minute
10                  Second
=================   ==============================================

CSEP (http://www.cseptesting.org) defines an extension to the format to
include uncertainties. The extended CSEP format has the following extra
columns:

=================   ==============================================
Column #            Value
=================   ==============================================
11                  Horizontal error
12                  Depth error
13                  Magnitude error
=================   ==============================================
"""
if __name__ == '__main__':
    import doctest

    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
Some Seismogram Interpolating Functions.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import numpy as np
import scipy.interpolate

from obspy.signal.headers import clibsignal


def _validate_parameters(data, old_start, old_dt, new_start, new_dt, new_npts):
    """
    Validates the parameters for various interpolation functions.

    Returns the old and the new end.
    """
    if new_dt <= 0.0:
        raise ValueError("The time step must be positive.")

    # Check for 1D array.
    if data.ndim != 1 or not len(data) or not data.shape[-1]:
        raise ValueError("Not a 1D array.")

    old_end = old_start + old_dt * (len(data) - 1)
    new_end = new_start + new_dt * (new_npts - 1)

    if old_start > new_start or old_end < new_end:
        raise ValueError("The new array must be fully contained in the old "
                         "array. No extrapolation can be performed.")

    return old_end, new_end


def interpolate_1d(data, old_start, old_dt, new_start, new_dt, new_npts,
                   type="linear", *args, **kwargs):
    """
    Wrapper around some scipy interpolation functions.

    :type data: array_like
    :param data: Array to interpolate.
    :type old_start: float
    :param old_start: The start of the array as a number.
    :type old_start: float
    :param old_dt: The time delta of the current array.
    :type new_start: float
    :param new_start: The start of the interpolated array. Must be greater
        or equal to the current start of the array.
    :type new_dt: float
    :param new_dt: The desired new time delta.
    :type new_npts: int
    :param new_npts: The new number of samples.
    :type type: str or int
    :param type: Specifies the kind of interpolation as a string (``linear``,
        ``nearest``, ``zero``, ``slinear``, ``quadratic``, ``cubic`` where
        ``slinear``, ``quadratic`` and ``cubic`` refer to a spline
        interpolation of first,  second or third order) or as an integer
        specifying the order of the spline interpolator to use. Default is
        ``linear``.
    """
    old_end, new_end = _validate_parameters(data, old_start, old_dt,
                                            new_start, new_dt, new_npts)

    # In almost all cases the unit will be in time.
    new_time_array = np.linspace(new_start, new_end, new_npts)
    old_time_array = np.linspace(old_start, old_end, len(data))

    s_map = {
        "slinear": 1,
        "quadratic": 2,
        "cubic": 3
    }
    if type in s_map:
        type = s_map[type]

    # InterpolatedUnivariateSpline uses a sane amount of memory for splines.
    # interp1d can easily require 50 GB of memory which is clearly
    # not acceptable.
    if isinstance(type, int):
        new_data = scipy.interpolate.InterpolatedUnivariateSpline(
            old_time_array, data, k=type)(new_time_array)
    # interp1d is used for the "linear", "nearest", and "zero" interpolation
    # methods.
    else:
        new_data = scipy.interpolate.interp1d(old_time_array, data, kind=type)(
            new_time_array)

    return new_data


def weighted_average_slopes(data, old_start, old_dt, new_start, new_dt,
                            new_npts, *args, **kwargs):
    r"""
    Implements the weighted average slopes interpolation scheme proposed in
    [Wiggins1976]_ for evenly sampled data. The scheme guarantees that there
    will be no additional extrema after the interpolation in contrast to
    spline interpolation.

    The slope :math:`s_i` at each knot is given by a weighted average of the
    adjacent linear slopes :math:`m_i` and :math:`m_{i+j}`:

    .. math::

        s_i = (w_i m_i + w_{i+1} m_{i+1}) / (w_i + w_{i+1})

    where

    .. math::

        w = 1 / max \left\{ \left\| m_i \\right\|, \epsilon \\right\}

    The value at each data point and the slope are then plugged into a
    piecewise continuous cubic polynomial used to evaluate the interpolated
    sample points.

    :type data: array_like
    :param data: Array to interpolate.
    :type old_start: float
    :param old_start: The start of the array as a number.
    :type old_start: float
    :param old_dt: The time delta of the current array.
    :type new_start: float
    :param new_start: The start of the interpolated array. Must be greater
        or equal to the current start of the array.
    :type new_dt: float
    :param new_dt: The desired new time delta.
    :type new_npts: int
    :param new_npts: The new number of samples.
    """
    old_end, new_end = _validate_parameters(data, old_start, old_dt,
                                            new_start, new_dt, new_npts)
    # In almost all cases the unit will be in time.
    new_time_array = np.linspace(new_start, new_end, new_npts)

    m = np.diff(data) / old_dt
    w = np.abs(m)
    w = 1.0 / np.clip(w, np.spacing(1), w.max())

    slope = np.empty(len(data), dtype=np.float64)
    slope[0] = m[0]
    slope[1:-1] = (w[:-1] * m[:-1] + w[1:] * m[1:]) / (w[:-1] + w[1:])
    slope[-1] = m[-1]

    # If m_i and m_{i+1} have opposite signs then set the slope to zero.
    # This forces the curve to have extrema at the sample points and not
    # in-between.
    sign_change = np.diff(np.sign(m)).astype(bool)
    slope[1:-1][sign_change] = 0.0

    derivatives = np.empty((len(data), 2), dtype=np.float64)
    derivatives[:, 0] = data
    derivatives[:, 1] = slope

    # Create interpolated value using hermite interpolation. In this case
    # it is directly applicable as the first derivatives are known.
    # Using scipy.interpolate.piecewise_polynomial_interpolate() is too
    # memory intensive
    return_data = np.empty(len(new_time_array), dtype=np.float64)
    clibsignal.hermite_interpolation(data, slope, new_time_array, return_data,
                                     len(data), len(return_data), old_dt,
                                     old_start)
    return return_data


# Map corresponding to the enum on the C side of things.
_LANCZOS_KERNEL_MAP = {
    "lanczos": 0,
    "hanning": 1,
    "blackman": 2
}


def lanczos_interpolation(data, old_start, old_dt, new_start, new_dt, new_npts,
                          a, window="lanczos", *args, **kwargs):
    r"""
    Function performing Lanczos resampling, see
    https://en.wikipedia.org/wiki/Lanczos_resampling for details. Essentially a
    finite support version of sinc resampling (the ideal reconstruction
    filter). For large values of ``a`` it converges towards sinc resampling. If
    used for downsampling, make sure to apply an appropriate anti-aliasing
    lowpass filter first.

    .. note::

        In most cases you do not want to call this method directly but invoke
        it via either the :meth:`obspy.core.stream.Stream.interpolate` or
        :meth:`obspy.core.trace.Trace.interpolate` method. These offer a nicer
        API that naturally integrates with the rest of ObsPy. Use
        ``method="lanczos"`` to use this interpolation method. In that case the
        only additional parameters of interest are ``a`` and ``window``.

    :type data: array_like
    :param data: Array to interpolate.
    :type old_start: float
    :param old_start: The start of the array as a number.
    :type old_start: float
    :param old_dt: The time delta of the current array.
    :type new_start: float
    :param new_start: The start of the interpolated array. Must be greater
        or equal to the current start of the array.
    :type new_dt: float
    :param new_dt: The desired new time delta.
    :type new_npts: int
    :param new_npts: The new number of samples.
    :type a: int
    :param a: The width of the window in samples on either side. Runtime
        scales linearly with the value of ``a`` but the interpolation also gets
        better.
    :type window: str
    :param window: The window used to taper the sinc function. One of
        ``"lanczos"``, ``"hanning"``, ``"blackman"``. The window determines
        the trade-off between "sharpness" and the amplitude of the wiggles in
        the pass and stop band. Please use the
        :func:`~obspy.signal.interpolation.plot_lanczos_windows` function to
        judge these for any given application.

    Values of ``a`` >= 20 show good results even for data that has
    energy close to the Nyquist frequency. If your data is extremely
    oversampled you can get away with much smaller ``a``'s.

    To get an idea of the response of the filter and the effect of the
    different windows, please use the
    :func:`~obspy.signal.interpolation.plot_lanczos_windows` function.

    Also be aware of any boundary effects. All values outside the data
    range are assumed to be zero which matters when calculating interpolated
    values at the boundaries. At each side the area with potential boundary
    effects is ``a`` * ``old_dt``. If you want to avoid any boundary effects
    you will have to remove these values.

    **Mathematical Details:**

    The :math:`\operatorname{sinc}` function is defined as

    .. math::

        \operatorname{sinc}(t) = \frac{\sin(\pi t)}{\pi t}.

    The Lanczos kernel is then given by a multiplication of the
    :math:`\operatorname{sinc}` function with an additional window function
    resulting in a finite support kernel.

    .. math::

        \begin{align}
            L(t) =
            \begin{cases}
                \operatorname{sinc}(t)\, \cdot \operatorname{sinc}(t/a)
                    & \text{if } t \in [-a, a]
                    \text{ and } \texttt{window} = \texttt{lanczos}\\
                \operatorname{sinc}(t)\, \cdot \frac{1}{2}
                (1 + \cos(\pi\, t/a))
                    & \text{if } t \in [-a, a]
                    \text{ and } \texttt{window} = \texttt{hanning}\\
                \operatorname{sinc}(t)\, \cdot \left( \frac{21}{50} +
                \frac{1}{2}
                \cos(\pi\, t/a) + \frac{2}{25} \cos (2\pi\, t/a) \right)
                    & \text{if } t \in [-a, a]
                    \text{ and } \texttt{window} = \texttt{blackman}\\
                0                     & \text{else}
            \end{cases}
        \end{align}


    Finally interpolation is performed by convolving the discrete signal
    :math:`s_i` with that kernel and evaluating it at the new time samples
    :math:`t_j`:

    .. math::

        \begin{align}
            S(t_j) =
                \sum_{i = \left \lfloor{t_j / \Delta t}\right \rfloor -a + 1}
                    ^{\left \lfloor{t_j / \Delta t}\right \rfloor + a}
            s_i L(t_j/\Delta t - i),
        \end{align}

    where :math:`\lfloor \cdot \rfloor` denotes the floor function. For more
    details and justification please see [Burger2009]_ and [vanDriel2015]_.
    """
    _validate_parameters(data, old_start, old_dt, new_start, new_dt, new_npts)

    # dt and offset in terms of the original sampling interval.
    dt_factor = float(new_dt) / old_dt
    offset = (new_start - old_start) / float(old_dt)

    if offset < 0:
        raise ValueError("Cannot extrapolate. Make sure to only interpolate "
                         "within the time range of the original signal.")

    if a < 1:
        raise ValueError("a must be at least 1.")

    return_data = np.zeros(new_npts, dtype=np.float64)

    clibsignal.lanczos_resample(
        np.require(data, dtype=np.float64), return_data, dt_factor, offset,
        len(data), len(return_data), int(a), 0)
    return return_data


def calculate_lanczos_kernel(x, a, window):
    """
    Helper function to get the actually used kernel for a specific value of
    a. Useful to analyse the behaviour of different tapers and different values
    of a.

    :type x: :class:`numpy.ndarray`
    :param x: The x values at which to calculate the kernel.
    :type a: int
    :param a: The width of the window in samples on either side.
    :type window: str
    :param window: The window used to multiply the sinc function with. One
        of ``"lanczos"``, ``"hanning"``, ``"blackman"``.

    Returns a dictionary of arrays:

    * ``"full_kernel"``: The tapered sinc function evaluated at samples ``x``.
    * ``"only_sinc"``: The sinc function evaluated at samples ``x``.
    * ``"only_taper"``: The taper function evaluated at samples ``x``.
    """
    window = window.lower()
    if window not in _LANCZOS_KERNEL_MAP:
        msg = "Invalid window. Valid windows: %s" % ", ".join(
            sorted(_LANCZOS_KERNEL_MAP.keys()))
        raise ValueError(msg)

    x = np.require(x, dtype=np.float64)
    y0 = np.zeros(x.shape, dtype=np.float64)
    y1 = np.zeros(x.shape, dtype=np.float64)
    y2 = np.zeros(x.shape, dtype=np.float64)

    clibsignal.calculate_kernel(
        x, y0, len(x), a, 0, _LANCZOS_KERNEL_MAP[window])
    clibsignal.calculate_kernel(
        x, y1, len(x), a, 1, _LANCZOS_KERNEL_MAP[window])
    clibsignal.calculate_kernel(
        x, y2, len(x), a, 2, _LANCZOS_KERNEL_MAP[window])

    ret_val = {
        "full_kernel": y0,
        "only_sinc": y1,
        "only_taper": y2
    }

    return ret_val


def plot_lanczos_windows(a, filename=None):
    """
    Helper function producing a plot of all available tapers of the sinc
    function and their response for the Lanczos interpolation.

    :type a: int
    :param a: The width of the window in samples on either side.
    :type filename: str
    :param filename: The file name to save to. The plot will be shown if it is
        not given.

    .. code-block:: python

        from obspy.signal.interpolation import plot_lanczos_windows
        plot_lanczos_windows(a=20)

    .. plot::

        import matplotlib.pyplot as plt
        plt.figure(figsize=(10, 15))
        from obspy.signal.interpolation import plot_lanczos_windows
        plot_lanczos_windows(a=20)
    """
    import matplotlib.pyplot as plt
    x_max = 1024.0 - 0.5
    n = 2 ** 15
    x = np.linspace(-x_max, x_max, n)
    dx = 2 * x_max / (n - 1)

    arrays = {}
    for key in _LANCZOS_KERNEL_MAP.keys():
        arrays[key] = calculate_lanczos_kernel(x, a, key)
        arrays[key]["fft"] = \
            np.abs(np.fft.rfft(arrays[key]["full_kernel"]) * dx)

    height = len(_LANCZOS_KERNEL_MAP) + 1

    plt.subplot(height, 2, 1)
    for key in sorted(arrays.keys()):
        plt.plot(x, arrays[key]["full_kernel"], label=key.capitalize())

    plt.legend()
    plt.xlim(-a, a)
    plt.ylim(-0.3, 1.1)
    plt.title("All Windows")

    plt.subplot(height, 2, 2)
    plt.plot([0.0, 0.5, 0.5, 1000], [1.0, 1.0, 0.0, 0.0], "--",
             color="0.1", label="Ideal")
    for key in sorted(arrays.keys()):
        plt.plot(np.abs(np.fft.fftfreq(len(x), dx)[:len(x) // 2 + 1]),
                 arrays[key]["fft"], label=key.capitalize())
    plt.xlim(0.2, 0.8)
    plt.ylim(-0.1, 1.1)
    plt.legend()
    plt.title("Frequency Response of All Windows")

    for _i, key in enumerate(sorted(_LANCZOS_KERNEL_MAP.keys())):
        plt.subplot(height, 2, 3 + 2 * _i)
        plt.title(key.capitalize())
        plt.plot(x, arrays[key]["full_kernel"], color="black", label="Final")
        plt.plot(x, arrays[key]["only_sinc"], "--", color="gray", label="Sinc")
        plt.plot(x, arrays[key]["only_taper"], color="red", label="Taper")
        plt.legend()
        plt.xlim(-a, a)
        plt.ylim(-0.3, 1.1)

        plt.subplot(height, 2, 3 + 2 * _i + 1)
        plt.title(key.capitalize() + " Response")
        plt.plot([0.0, 0.5, 0.5, 1000], [1.0, 1.0, 0.0, 0.0], "--",
                 color="0.1")
        plt.plot(np.abs(np.fft.fftfreq(len(x), dx)[:len(x) // 2 + 1]),
                 arrays[key]["fft"])
        plt.xlim(0.2, 0.8)
        plt.ylim(-0.1, 1.1)

    plt.suptitle("Different windows for sinc interpolation with a=%i"
                 % a, fontsize="large")

    if not filename:
        plt.show()
    else:
        plt.savefig(filename)
        plt.close()



# -*- coding: utf-8 -*-
import pytest

from obspy import UTCDateTime
from obspy.core.inventory import Comment
from obspy.core.util import (ComplexWithUncertainties, Enum,
                             FloatWithUncertainties)
from obspy.core.util.obspy_types import (FloatWithUncertaintiesAndUnit)


class TestUtilTypes:
    """
    Test suite for obspy.core.util.base
    """
    def test_enum(self):
        """
        Tests for the enum type.
        """
        items = ["m", "s", "m/s", "m/(s*s)", "m*s", "dimensionless", "other"]
        units = Enum(items)
        # existing selections
        assert units.other == "other"
        assert units.M == "m"
        assert units['m/s'] == "m/s"
        assert units.get('m/s') == "m/s"
        assert units[0] == "m"
        assert units[-1] == "other"
        # not existing selections should fail
        with pytest.raises(Exception):
            units.__getitem__('5')
        with pytest.raises(Exception):
            units.__getattr__('xx')
        with pytest.raises(Exception):
            units.get('xx', 'default')
        with pytest.raises(Exception):
            units.__getitem__(99)
        with pytest.raises(Exception):
            units.__getitem__(-99)
        # test in operator
        assert "other" in units
        assert "ot21her" not in units
        # test typical dict methods
        assert units.values() == items
        assert units.items() == list(zip(items, items))
        assert units.keys() == items
        # call will either return correct enum label or return None
        assert units('m') == 'm'
        assert units('m/(s*s)') == 'm/(s*s)'
        assert units(5) == 'dimensionless'
        assert units(99) is None
        assert units('xxx') is None

    def _check_complex_with_u(self, c, real, r_lo, r_up, imag, i_lo, i_up):
        """
        Check for six equalities for a ComplexWithUncertainties
        """
        assert isinstance(c.real, FloatWithUncertainties)
        assert isinstance(c.imag, FloatWithUncertainties)
        assert c.real == real
        assert c.imag == imag
        assert c.real.upper_uncertainty == r_up
        assert c.real.lower_uncertainty == r_lo
        assert c.imag.upper_uncertainty == i_up
        assert c.imag.lower_uncertainty == i_lo

    def test_complex(self):
        """
        Test the ComplexWithUncertainties for proper usage
        """
        f1 = float(3.5)
        f2 = float(12)
        lu1 = 1
        uu1 = 2
        lu2 = 4.1
        uu2 = 7.2
        fu1 = FloatWithUncertainties(f1, lower_uncertainty=lu1,
                                     upper_uncertainty=uu1)
        fu2 = FloatWithUncertainties(f2, lower_uncertainty=lu2,
                                     upper_uncertainty=uu2)
        c1 = ComplexWithUncertainties()
        c2 = ComplexWithUncertainties(f1, f2)
        c3 = ComplexWithUncertainties(
            f1, f2, lower_uncertainty=complex(lu1, lu2),
            upper_uncertainty=complex(uu1, uu2))
        c4 = ComplexWithUncertainties(fu1, fu2)
        # c1 should be 0+0j with uncertainties of None
        self._check_complex_with_u(c1, 0, None, None, 0, None, None)
        # c2 should return the floats
        self._check_complex_with_u(c2, f1, None, None, f2, None, None)
        # c3 and c4 should be the same.
        self._check_complex_with_u(c3, f1, lu1, uu1, f2, lu2, uu2)
        self._check_complex_with_u(c4, f1, lu1, uu1, f2, lu2, uu2)
        assert c4.real == fu1
        assert c4.imag == fu2

    def test_floating_point_types_are_indempotent(self):
        """
        Applying the constructor multiple times should not change the values.
        """
        f = FloatWithUncertainties(1.0, lower_uncertainty=0.5,
                                   upper_uncertainty=1.5)
        assert f == 1.0
        assert f.lower_uncertainty == 0.5
        assert f.upper_uncertainty == 1.5
        f = FloatWithUncertainties(f)
        assert f == 1.0
        assert f.lower_uncertainty == 0.5
        assert f.upper_uncertainty == 1.5

        f = FloatWithUncertaintiesAndUnit(1.0, lower_uncertainty=0.5,
                                          upper_uncertainty=1.5, unit="AB")
        assert f == 1.0
        assert f.lower_uncertainty == 0.5
        assert f.upper_uncertainty == 1.5
        assert f.unit == "AB"
        f = FloatWithUncertaintiesAndUnit(f)
        assert f == 1.0
        assert f.lower_uncertainty == 0.5
        assert f.upper_uncertainty == 1.5
        assert f.unit == "AB"

    def test_comment_str(self):
        """
        Tests the __str__ method of the Comment object.
        """
        c = Comment(value='test_comment', id=9,
                    begin_effective_time=UTCDateTime(1240561632),
                    end_effective_time=UTCDateTime(1584561632), authors=[])
        cs = str(c)
        assert cs == "Comment:\ttest_comment\n" \
                     "\tBegin Effective Time:\t2009-04-24T08:27:12" \
                     ".000000Z\n" \
                     "\tEnd Effective Time:\t2020-03-18T20:00:32.000000Z\n" \
                     "\tAuthors:\t\t[]\n" \
                     "\tId:\t\t\t9"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Functions for testing the obspy.io.nordic functions
"""
import io
import os
import warnings
from itertools import cycle
import numpy as np

from obspy import read_events, Catalog, UTCDateTime, read
from obspy.core.event import (
    Pick, WaveformStreamID, Arrival, Amplitude, Event, Origin, Magnitude,
    OriginQuality, EventDescription, CreationInfo, OriginUncertainty,
    ConfidenceEllipsoid, QuantityError, FocalMechanism, MomentTensor,
    NodalPlane, NodalPlanes, ResourceIdentifier, Tensor)
from obspy.core.util.base import NamedTemporaryFile
from obspy.core.util.misc import TemporaryWorkingDirectory

from obspy.io.nordic import NordicParsingError
from obspy.io.nordic.core import (
    _is_sfile, read_spectral_info, read_nordic, readwavename, blanksfile,
    _write_nordic, nordpick, readheader, _readheader, write_select)
from obspy.io.nordic.utils import (
    _int_conv, _float_conv, _str_conv, _nortoevmag, _evmagtonor,
    _get_line_tags)
from obspy.io.nordic.ellipse import Ellipse
import pytest


class TestNordicMethods:
    """
    Test suite for nordic io operations.
    """
    def test_read_write(self):
        """
        Function to test the read and write capabilities of sfile_util.
        """
        # Set-up a test event
        test_event = full_test_event()
        # Sort the magnitudes - they are sorted on writing and we need to check
        # like-for-like
        test_event.magnitudes.sort(key=lambda obj: obj['mag'], reverse=True)
        # Add the event to a catalogue which can be used for QuakeML testing
        test_cat = Catalog()
        test_cat += test_event
        for nordic_format in ['OLD', 'NEW']:
            # Check the read-write s-file functionality
            with TemporaryWorkingDirectory():
                with warnings.catch_warnings():
                    # Evaluation mode mapping warning
                    warnings.simplefilter('ignore', UserWarning)
                    sfile = _write_nordic(
                        test_cat[0], filename=None, userid='TEST', evtype='L',
                        outdir='.', wavefiles='test', explosion=True,
                        overwrite=True, nordic_format=nordic_format)
                assert readwavename(sfile) == ['test']
                read_cat = Catalog()
                # raises "UserWarning: AIN in header, currently unsupported"
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', UserWarning)
                    read_cat += read_nordic(sfile)
            read_ev = read_cat[0]
            test_ev = test_cat[0]
            for read_pick, test_pick in zip(read_ev.picks, test_ev.picks):
                assert read_pick.time == test_pick.time
                assert read_pick.backazimuth == test_pick.backazimuth
                assert read_pick.onset == test_pick.onset
                assert read_pick.phase_hint == test_pick.phase_hint
                if test_pick.polarity == "undecidable":
                    assert read_pick.polarity is None
                elif read_pick.polarity == "undecidable":
                    assert test_pick.polarity is None
                else:
                    assert read_pick.polarity == test_pick.polarity
                assert read_pick.waveform_id.station_code == \
                    test_pick.waveform_id.station_code
                assert read_pick.waveform_id.channel_code[-1] == \
                    test_pick.waveform_id.channel_code[-1]
            # assert read_ev.origins[0].resource_id ==\
            #     test_ev.origins[0].resource_id
            assert read_ev.origins[0].time == \
                test_ev.origins[0].time
            # Note that time_residual_RMS is not a quakeML format
            assert read_ev.origins[0].longitude == \
                test_ev.origins[0].longitude
            assert read_ev.origins[0].latitude == \
                test_ev.origins[0].latitude
            assert read_ev.origins[0].depth == \
                test_ev.origins[0].depth
            assert read_ev.magnitudes[0].mag == \
                test_ev.magnitudes[0].mag
            assert read_ev.magnitudes[1].mag == \
                test_ev.magnitudes[1].mag
            assert read_ev.magnitudes[2].mag == \
                test_ev.magnitudes[2].mag
            assert read_ev.magnitudes[0].creation_info == \
                test_ev.magnitudes[0].creation_info
            assert read_ev.magnitudes[1].creation_info == \
                test_ev.magnitudes[1].creation_info
            assert read_ev.magnitudes[2].creation_info == \
                test_ev.magnitudes[2].creation_info
            assert read_ev.magnitudes[0].magnitude_type == \
                test_ev.magnitudes[0].magnitude_type
            assert read_ev.magnitudes[1].magnitude_type == \
                test_ev.magnitudes[1].magnitude_type
            assert read_ev.magnitudes[2].magnitude_type == \
                test_ev.magnitudes[2].magnitude_type
            assert read_ev.event_descriptions == \
                test_ev.event_descriptions
            # assert read_ev.amplitudes[0].resource_id ==\
            #     test_ev.amplitudes[0].resource_id
            assert read_ev.amplitudes[0].period == \
                test_ev.amplitudes[0].period
            assert read_ev.amplitudes[0].snr == \
                test_ev.amplitudes[0].snr
            assert read_ev.amplitudes[2].period == \
                test_ev.amplitudes[2].period
            assert read_ev.amplitudes[2].snr == \
                test_ev.amplitudes[2].snr
            # Check coda magnitude pick
            # Resource ids get overwritten because you can't have two the same
            # in memory
            # self.assertEqual(read_ev.amplitudes[1].resource_id,
            #                  test_ev.amplitudes[1].resource_id)
            assert read_ev.amplitudes[1].type == \
                test_ev.amplitudes[1].type
            assert read_ev.amplitudes[1].unit == \
                test_ev.amplitudes[1].unit
            assert read_ev.amplitudes[1].generic_amplitude == \
                test_ev.amplitudes[1].generic_amplitude
            # Resource ids get overwritten because you can't have two the same
            # in memory
            # self.assertEqual(read_ev.amplitudes[1].pick_id,
            #                  test_ev.amplitudes[1].pick_id)
            assert read_ev.amplitudes[1].waveform_id.station_code == \
                test_ev.amplitudes[1].waveform_id.station_code
            if nordic_format == 'OLD':
                assert read_ev.amplitudes[1].waveform_id.channel_code == \
                    test_ev.amplitudes[1].waveform_id.channel_code[0] + \
                    test_ev.amplitudes[1].waveform_id.channel_code[-1]
            assert read_ev.amplitudes[1].magnitude_hint == \
                test_ev.amplitudes[1].magnitude_hint
            # snr is not supported in s-file
            # self.assertEqual(read_ev.amplitudes[1].snr,
            #                  test_ev.amplitudes[1].snr)
            assert read_ev.amplitudes[1].category == \
                test_ev.amplitudes[1].category
            assert read_ev.extra.get('nordic_event_id')['value'] == \
                test_ev.extra.get('nordic_event_id')['value']

    def test_write_read_quakeml(self, testdata):
        """
        Test whether all properties are set properly such that file can be
        written to Quakeml and read back in.
        """
        testing_path = testdata['03-0345-23L.S202101']
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_nordic(testing_path)
        with NamedTemporaryFile(suffix='.qml') as tf:
            cat.write(tf.name, format='QUAKEML')
            cat_back = read_events(tf.name)
            for event_1, event_2 in zip(cat, cat_back):
                _assert_similarity(event_1=event_1, event_2=event_2,
                                   strict=False)
        with NamedTemporaryFile(suffix='.out') as tf:
            cat.write(tf.name, format='NORDIC', nordic_format='NEW')
            cat_back = read_events(tf.name)
            for event_1, event_2 in zip(cat, cat_back):
                _assert_similarity(event_1=event_1, event_2=event_2,
                                   strict=False)

    def test_read_compact_file(self, testdata):
        """
        Test reading of a compact Nordic file with only header lines.
        """
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_nordic(testdata['collect.out'])
        assert len(cat) == 3
        for event in cat:
            assert len(event.picks) == 0

    def test_fail_writing(self):
        """
        Test a deliberate fail.
        """
        test_event = full_test_event()
        # Add the event to a catalogue which can be used for QuakeML testing
        test_cat = Catalog()
        test_cat += test_event
        test_ev = test_cat[0]
        test_cat.append(full_test_event())
        with pytest.raises(NordicParsingError):
            # Raises error due to multiple events in catalog
            with warnings.catch_warnings():
                # Evaluation mode mapping warning
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(test_cat, filename=None, userid='TEST',
                              evtype='L', outdir='.',
                              wavefiles='test', explosion=True,
                              overwrite=True)
        with pytest.raises(NordicParsingError):
            # Raises error due to too long userid
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(test_ev, filename=None, userid='TESTICLE',
                              evtype='L', outdir='.',
                              wavefiles='test', explosion=True,
                              overwrite=True)
        with pytest.raises(NordicParsingError):
            # Raises error due to unrecognised event type
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(test_ev, filename=None, userid='TEST',
                              evtype='U', outdir='.',
                              wavefiles='test', explosion=True,
                              overwrite=True)
        with pytest.raises(NordicParsingError):
            # Raises error due to no output directory
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(test_ev, filename=None, userid='TEST',
                              evtype='L', outdir='albatross',
                              wavefiles='test', explosion=True,
                              overwrite=True)
        invalid_origin = test_ev.copy()

        invalid_origin.origins = []
        with pytest.raises(NordicParsingError):
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(invalid_origin, filename=None, userid='TEST',
                              evtype='L', outdir='.', wavefiles='test',
                              explosion=True, overwrite=True)
        invalid_origin = test_ev.copy()
        invalid_origin.origins[0].time = None
        with pytest.raises(NordicParsingError):
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(invalid_origin, filename=None, userid='TEST',
                              evtype='L', outdir='.', wavefiles='test',
                              explosion=True, overwrite=True)
        # Write a near empty origin
        valid_origin = test_ev.copy()
        valid_origin.origins[0].latitude = None
        valid_origin.origins[0].longitude = None
        valid_origin.origins[0].depth = None
        with NamedTemporaryFile() as tf:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                _write_nordic(
                    valid_origin, filename=tf.name, userid='TEST',
                    evtype='L', outdir='.', wavefiles='test',
                    explosion=True, overwrite=True)
            assert os.path.isfile(tf.name)

    def test_blanksfile(self):
        st = read()
        with TemporaryWorkingDirectory():
            testing_path = 'Temporary_wavefile'
            st.write(testing_path, format='MSEED')
            sfile = blanksfile(testing_path, 'L', 'TEST', overwrite=True)
            assert os.path.isfile(sfile)
            os.remove(sfile)
            sfile = blanksfile(testing_path, 'L', 'TEST', overwrite=True,
                               evtime=UTCDateTime())
            assert os.path.isfile(sfile)
            os.remove(sfile)
            with pytest.raises(NordicParsingError):
                # No wavefile
                blanksfile('albert', 'L', 'TEST', overwrite=True)
            with pytest.raises(NordicParsingError):
                # USER ID too long
                blanksfile(testing_path, 'L', 'TESTICLE', overwrite=True)
            with pytest.raises(NordicParsingError):
                # Unknown event type
                blanksfile(testing_path, 'U', 'TEST', overwrite=True)
            # Check that it breaks when writing multiple versions
            sfiles = []
            for _i in range(10):
                # raises UserWarning: Desired sfile exists, will not overwrite
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', UserWarning)
                    sfiles.append(blanksfile(testing_path, 'L', 'TEST'))
            with pytest.raises(NordicParsingError):
                # raises UserWarning: Desired sfile exists, will not overwrite
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', UserWarning)
                    blanksfile(testing_path, 'L', 'TEST')
            for sfile in sfiles:
                assert os.path.isfile(sfile)

    def test_write_empty(self):
        """
        Function to check that writing a blank event works as it should.
        """
        test_event = Event()
        with pytest.raises(NordicParsingError):
            _write_nordic(test_event, filename=None, userid='TEST', evtype='L',
                          outdir='.', wavefiles='test')
        test_event.origins.append(Origin())
        with pytest.raises(NordicParsingError):
            _write_nordic(test_event, filename=None, userid='TEST', evtype='L',
                          outdir='.', wavefiles='test')
        test_event.origins[0].time = UTCDateTime()
        with TemporaryWorkingDirectory():
            test_sfile = _write_nordic(test_event, filename=None,
                                       userid='TEST', evtype='L', outdir='.',
                                       wavefiles='test')
            assert os.path.isfile(test_sfile)

    def test_read_empty_header(self, testdata):
        """
        Function to check a known issue, empty header info S-file: Bug found \
        by Dominic Evanzia.
        """
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)

            test_event = read_nordic(testdata['Sfile_no_location'])[0]
        assert not test_event.origins[0].latitude
        assert not test_event.origins[0].longitude
        assert not test_event.origins[0].depth

    def test_read_extra_header(self, testdata):
        testing_path = testdata['Sfile_extra_header']
        not_extra_header = testdata['01-0411-15L.S201309']
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            test_event = read_nordic(testing_path)[0]
            header_event = read_nordic(not_extra_header)[0]
        assert len(header_event.origins) == 2
        assert test_event.origins[0].time == \
               header_event.origins[0].time
        assert test_event.origins[0].latitude == \
               header_event.origins[0].latitude
        assert test_event.origins[0].longitude == \
               header_event.origins[0].longitude
        assert test_event.origins[0].depth == \
               header_event.origins[0].depth

    def test_header_mapping(self, testdata):
        # Raise "UserWarning: Lines of type I..."
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            head_1 = readheader(testdata['01-0411-15L.S201309'])
        with open(testdata['01-0411-15L.S201309'], 'r') as f:
            # raises "UserWarning: AIN in header, currently unsupported"
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                tagged_lines = _get_line_tags(f=f)
                head_2, _ = _readheader(head_lines=tagged_lines['1'])
        _assert_similarity(head_1, head_2, strict=True)

    def test_missing_header(self, testdata):
        # Check that a suitable error is raised
        with pytest.raises(NordicParsingError):
            # Raises AIN warning
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                readheader(testdata['Sfile_no_header'])

    def test_reading_string_io(self, testdata):
        filename = testdata['01-0411-15L.S201309']
        with open(filename, "rt") as fh:
            file_object = io.StringIO(fh.read())

        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)

            cat = read_events(file_object)
            file_object.close()

            ref_cat = read_events(filename)
            _assert_similarity(cat[0], ref_cat[0], strict=True)

    def test_reading_bytes_io(self, testdata):
        filename = testdata['01-0411-15L.S201309']
        with open(filename, "rb") as fh:
            file_object = io.BytesIO(fh.read())

        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)

            cat = read_events(file_object)
            file_object.close()

            ref_cat = read_events(filename)
            _assert_similarity(cat[0], ref_cat[0], strict=True)

    def test_corrupt_header(self, testdata):
        filename = testdata['01-0411-15L.S201309']
        f = open(filename, 'r')
        with NamedTemporaryFile(suffix='.sfile') as tmp_file:
            fout = open(tmp_file.name, 'w')
            for line in f:
                fout.write(line[0:78])
            f.close()
            fout.close()
            with pytest.raises(NordicParsingError):
                readheader(tmp_file.name)

    def test_multi_writing(self):
        event = full_test_event()
        # Try to write the same event multiple times, but not overwrite
        sfiles = []
        with TemporaryWorkingDirectory():
            for _i in range(59):
                with warnings.catch_warnings():
                    # Evaluation mode mapping warning
                    warnings.simplefilter('ignore', UserWarning)
                    sfiles.append(_write_nordic(event=event, filename=None,
                                                overwrite=False))
            with pytest.raises(NordicParsingError):
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', UserWarning)
                    _write_nordic(event=event, filename=None, overwrite=False)

    def test_mag_conv(self):
        """
        Check that we convert magnitudes as we should!
        """
        magnitude_map = [
            ('L', 'ML'), ('B', 'mB'), ('S', 'MS'), ('W', 'MW'), ('G', 'MbLg'),
            ('C', 'Mc'), ('s', 'Ms')]
        for magnitude in magnitude_map:
            assert magnitude[0] == _evmagtonor(magnitude[1])
            assert _nortoevmag(magnitude[0]) == magnitude[1]

    def test_str_conv(self):
        """
        Test the simple string conversions.
        """
        assert _int_conv('albert') is None
        assert _float_conv('albert') is None
        assert _str_conv('albert') == 'albert'
        assert _int_conv('1') == 1
        assert _float_conv('1') == 1.0
        assert _str_conv(1) == '1'
        assert _int_conv('1.0256') is None
        assert _float_conv('1.0256') == 1.0256
        assert _str_conv(1.0256) == '1.0256'

    def test_read_wavename(self, testdata):
        testing_path = testdata['01-0411-15L.S201309']
        wavefiles = readwavename(testing_path)
        assert len(wavefiles) == 1
        # Check that read_nordic reads wavname when return_wavnames=True
        cat, wavefiles = read_nordic(testing_path, return_wavnames=True)
        assert wavefiles == [['2013-09-01-0410-35.DFDPC_024_00']]
        # Test that full paths are handled
        test_event = full_test_event()
        # Add the event to a catalogue which can be used for QuakeML testing
        test_cat = Catalog()
        test_cat += test_event
        # Check the read-write s-file functionality
        with TemporaryWorkingDirectory():
            with warnings.catch_warnings():
                # Evaluation mode mapping warning
                warnings.simplefilter('ignore', UserWarning)
                sfile = _write_nordic(
                    test_cat[0], filename=None, userid='TEST', evtype='L',
                    outdir='.', wavefiles=['walrus/test'], explosion=True,
                    overwrite=True)
            assert readwavename(sfile) == ['test']
        # Check that multiple wavefiles are read properly
        with TemporaryWorkingDirectory():
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                sfile = _write_nordic(
                    test_cat[0], filename=None, userid='TEST', evtype='L',
                    outdir='.', wavefiles=['walrus/test', 'albert'],
                    explosion=True, overwrite=True)
            assert readwavename(sfile) == ['test', 'albert']

    def test_read_event(self, testdata):
        """
        Test the wrapper.
        """
        testing_path = testdata['01-0411-15L.S201309']
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            event = read_nordic(testing_path)[0]
        assert len(event.origins) == 2
        assert len(event.picks) == 17

    def test_read_event_new(self, testdata):
        """
        Test the wrapper for New Nordic format reading.
        """
        testing_path = testdata['03-0345-23L.S202101']
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            event = read_nordic(testing_path)[0]
        assert len(event.origins) == 1
        assert len(event.picks) == 53
        assert len(event.origins[0].arrivals) == 35

    def test_read_latin1(self, testdata):
        """
        Check that we can read dos formatted, latin1 encoded files.
        """
        with warnings.catch_warnings():
            # Lots of unsupported line warnings
            warnings.simplefilter('ignore', UserWarning)
            dos_file = testdata['dos-file.sfile']
            assert _is_sfile(dos_file)
            event = readheader(dos_file)
            assert event.origins[0].latitude == 60.328
            cat = read_events(dos_file)
            assert cat[0].origins[0].latitude == 60.328
            wavefiles = readwavename(dos_file)
            assert wavefiles[0] == "90121311.0851W41"
            spectral_info = read_spectral_info(dos_file)
            assert len(spectral_info.keys()) == 10
            assert spectral_info[('AVERAGE', '')]['stress_drop'] == 27.7
            with pytest.raises(UnicodeDecodeError):
                readheader(dos_file, 'ASCII')

    def test_read_many_events(self, testdata):
        testing_path = testdata['select.out']
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            catalog = read_nordic(testing_path)
        assert len(catalog) == 50

    def test_write_select(self):
        cat = read_events()
        with NamedTemporaryFile(suffix='.out') as tf:
            # raises "UserWarning: mb is not convertible"
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                write_select(cat, filename=tf.name, nordic_format='OLD')
            assert _is_sfile(tf.name)
            with warnings.catch_warnings():
                # Type I warning
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
        for event_1, event_2 in zip(cat, cat_back):
            _assert_similarity(event_1=event_1, event_2=event_2, strict=False)

    def test_write_select_new(self):
        cat = read_events()
        with NamedTemporaryFile(suffix='.out') as tf:
            # raises "UserWarning: mb is not convertible"
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                write_select(cat, filename=tf.name, nordic_format='NEW')
            assert _is_sfile(tf.name)
            with warnings.catch_warnings():
                # Type I warning
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
        for event_1, event_2 in zip(cat, cat_back):
            _assert_similarity(event_1=event_1, event_2=event_2, strict=False)

    def test_write_plugin(self):
        cat = read_events()
        cat.append(full_test_event())
        with NamedTemporaryFile(suffix='.out') as tf:
            # raises UserWarning: mb is not convertible
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat.write(tf.name, format='nordic')
            # raises "UserWarning: AIN in header, currently unsupported"
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
            for event_1, event_2 in zip(cat, cat_back):
                _assert_similarity(event_1=event_1, event_2=event_2,
                                   strict=False)

    def test_more_than_three_mags(self):
        cat = Catalog()
        cat += full_test_event()
        cat[0].magnitudes.append(Magnitude(
            mag=0.9, magnitude_type='MS', creation_info=CreationInfo('TES'),
            origin_id=cat[0].origins[0].resource_id))
        with NamedTemporaryFile(suffix='.out') as tf:
            # raises UserWarning: mb is not convertible
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat.write(tf.name, format='nordic')
            # raises "UserWarning: AIN in header, currently unsupported"
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
            for event_1, event_2 in zip(cat, cat_back):
                assert len(event_1.magnitudes) == len(event_2.magnitudes)
                _assert_similarity(event_1, event_2, strict=False)

    def test_inaccurate_picks(self, testdata):
        testing_path = testdata['bad_picks.sfile']
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_nordic(testing_path)
        pick_string = nordpick(cat[0])
        for pick in pick_string:
            assert len(pick) == 80

    def test_round_len(self, testdata):
        testing_path = testdata['round_len_undef.sfile']
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            event = read_nordic(testing_path)[0]
        pick_string = nordpick(event)
        for pick in pick_string:
            assert len(pick) == 80

    def test_read_moment(self, testdata):
        """
        Test the reading of seismic moment from the s-file.
        """
        testing_path = testdata['automag.out']
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            event = read_nordic(testing_path)[0]
        mag = [m for m in event.magnitudes if m.magnitude_type == 'MW']
        assert len(mag) == 1
        assert mag[0].mag == 0.7

    def test_read_moment_info(self, testdata):
        """
        Test reading the info from spectral analysis.
        """
        testing_path = testdata['automag.out']
        with warnings.catch_warnings():
            # Userwarning, type I
            warnings.simplefilter('ignore', UserWarning)
            spec_inf = read_spectral_info(testing_path)
        assert len(spec_inf) == 5
        # This should actually test that what we are reading in is correct.
        average = spec_inf[('AVERAGE', '')]
        check_av = {u'channel': '', u'corner_freq': 5.97, u'decay': 0.0,
                    u'moment': 12589254117.941662, u'moment_mag': 0.7,
                    u'source_radius': 0.231,
                    u'spectral_level': 0.3981071705534972,
                    u'station': 'AVERAGE', u'stress_drop': 0.006,
                    u'window_length': 1.6}
        for key in average.keys():
            if isinstance(average.get(key), str):
                assert average.get(key) == check_av.get(key)
            else:
                assert round(average.get(key), 4) == \
                                 round(check_av.get(key), 4)

    def test_is_sfile(self, datapath, testdata):
        sfiles = ['01-0411-15L.S201309', 'automag.out', 'bad_picks.sfile',
                  'round_len_undef.sfile', 'Sfile_extra_header',
                  'Sfile_no_location']
        for sfile in sfiles:
            assert _is_sfile(testdata[sfile])
        no_header_path = testdata['Sfile_no_header']
        assert not _is_sfile(no_header_path)
        assert not _is_sfile(
            datapath.parent.parent.parent /
            'nlloc' / 'tests' / 'data' / 'nlloc.hyp')

    def test_read_picks_across_day_end(self, testdata):
        testing_path = testdata['sfile_over_day']
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            event = read_nordic(testing_path)[0]
        pick_times = [pick.time for pick in event.picks]
        for pick in event.picks:
            # Pick should come after origin time
            assert pick.time > event.origins[0].time
            # All picks in this event are within 60s of origin time
            assert (pick.time - event.origins[0].time) <= 60
        # Make sure zero hours and 24 hour picks are handled the same.
        testing_path = testdata['sfile_over_day_zeros']
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            event_2 = read_nordic(testing_path)[0]
        for pick in event_2.picks:
            # Pick should come after origin time
            assert pick.time > event_2.origins[0].time
            # All picks in this event are within 60s of origin time
            assert (pick.time - event_2.origins[0].time) <= 60
            # Each pick should be the same as one pick in the previous event
            assert pick.time in pick_times
        assert event_2.origins[0].time == event.origins[0].time

    def test_distance_conversion(self, testdata):
        """
        Check that distances are converted properly.
        """
        testing_path = testdata['01-0411-15L.S201309']
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(testing_path)
        event = cat[0]
        dist = sorted(event.origins[0].arrivals, key=lambda x: x.distance)[0]
        assert round(abs(dist.distance - 0.035972864236749225), 7) == 0
        pick_strings = nordpick(event)
        assert int([p for p in pick_strings if p.split()[0] == 'GCSZ' and
               p.split()[1] == 'SZ'][0].split()[-1]) == 304
        assert int([p for p in pick_strings if p.split()[0] == 'WZ11' and
               p.split()[1] == 'HZ'][0].split()[-1]) == 30

    def test_large_negative_longitude(self):
        event = full_test_event()
        event.origins[0].longitude = -120
        with NamedTemporaryFile(suffix=".out") as tf:
            with warnings.catch_warnings():
                # Evaluation mode mapping warning
                warnings.simplefilter('ignore', UserWarning)
                event.write(tf.name, format="NORDIC")
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                event_back = read_events(tf.name)
        _assert_similarity(event, event_back[0], strict=True)

    def test_write_preferred_origin(self):
        event = full_test_event()
        preferred_origin = Origin(
            time=UTCDateTime("2012-03-26") + 2.2, latitude=47.0,
            longitude=35.0, depth=18000)
        event.origins.append(preferred_origin)
        event.preferred_origin_id = preferred_origin.resource_id
        with NamedTemporaryFile(suffix=".out") as tf:
            with warnings.catch_warnings():
                # Evaluation mode mapping warning
                warnings.simplefilter('ignore', UserWarning)
                event.write(tf.name, format="NORDIC")
            with warnings.catch_warnings():
                # Type I warning
                warnings.simplefilter('ignore', UserWarning)
                event_back = read_events(tf.name)
        assert preferred_origin.latitude == event_back[0].origins[0].latitude
        assert preferred_origin.longitude == event_back[0].origins[0].longitude
        assert preferred_origin.depth == event_back[0].origins[0].depth
        assert preferred_origin.time == event_back[0].origins[0].time

    def test_read_high_precision_pick(self, testdata):
        """
        Nordic supports writing to milliseconds in high-precision mode,
        obspy < 1.2.0 did not properly read this, see #2348.
        """
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(
                testdata['sfile_high_precision_picks'])
        event = cat[0]
        pick_times = {
            "LSd1": UTCDateTime(2010, 11, 26, 1, 28, 46.859),
            "LSd3": UTCDateTime(2010, 11, 26, 1, 28, 48.132),
            "LSd2": UTCDateTime(2010, 11, 26, 1, 28, 48.183),
            "LSd4": UTCDateTime(2010, 11, 26, 1, 28, 49.744)}
        for key, value in pick_times.items():
            pick = [p for p in event.picks
                    if p.waveform_id.station_code == key]
            assert len(pick) == 1
            assert pick[0].time == value

    def test_high_precision_read_write(self, testdata):
        """ Test that high-precision writing works. """
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(testdata['sfile_high_precision_picks'])
        event = cat[0]
        pick_times = {
            "LSd1": UTCDateTime(2010, 11, 26, 1, 28, 46.859),
            "LSd3": UTCDateTime(2010, 11, 26, 1, 28, 48.132),
            "LSd2": UTCDateTime(2010, 11, 26, 1, 28, 48.183),
            "LSd4": UTCDateTime(2010, 11, 26, 1, 28, 49.744)}
        for key, value in pick_times.items():
            pick = [p for p in event.picks
                    if p.waveform_id.station_code == key]
            assert len(pick) == 1
            assert pick[0].time == value
        with NamedTemporaryFile(suffix=".out") as tf:
            write_select(cat, filename=tf.name, nordic_format='OLD')
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
        assert len(cat_back) == 1
        for key, value in pick_times.items():
            pick = [p for p in cat_back[0].picks
                    if p.waveform_id.station_code == key]
            assert len(pick) == 1
            assert pick[0].time == value
        # Check that writing to standard accuracy just gives a rounded version
        with NamedTemporaryFile(suffix=".out") as tf:
            cat.write(format="NORDIC", filename=tf.name, high_accuracy=False)
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
        assert len(cat_back) == 1
        for key, value in pick_times.items():
            pick = [p for p in cat_back[0].picks
                    if p.waveform_id.station_code == key]
            assert len(pick) == 1
            rounded_pick_time = UTCDateTime(
                value.year, value.month, value.day, value.hour, value.minute)
            rounded_pick_time += round(
                value.second + (value.microsecond / 1e6), 2)
            assert pick[0].time == rounded_pick_time

    def test_long_phase_name(self, testdata):
        """ Nordic format supports 8 char phase names, sometimes. """
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(testdata['sfile_long_phase'])
        # This file has one event with one pick
        pick = cat[0].picks[0]
        arrival = cat[0].origins[0].arrivals[0]
        assert pick.phase_hint == "PKiKP"
        assert arrival.time_weight == 1
        with NamedTemporaryFile(suffix=".out") as tf:
            write_select(cat, filename=tf.name, nordic_format='OLD')
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
        pick = cat_back[0].picks[0]
        arrival = cat_back[0].origins[0].arrivals[0]
        assert pick.phase_hint == "PKiKP"
        assert arrival.time_weight == 1

    def test_read_write_over_day(self):
        """
        Nordic picks are relative to origin time - check that this works
        over day boundaries.
        """
        event = full_test_event()
        event.origins[0].time -= 3600
        assert event.picks[0].time.date > event.origins[0].time.date
        with NamedTemporaryFile(suffix=".out") as tf:
            with warnings.catch_warnings():
                # Evaluation mode mapping warning
                warnings.simplefilter('ignore', UserWarning)
                write_select(Catalog([event]), filename=tf.name,
                             nordic_format='OLD')
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                event_back = read_events(tf.name)[0]
        _assert_similarity(event, event_back, strict=True)

    def test_seconds_overflow(self, testdata):
        """
        #2348 indicates that SEISAN sometimes overflows seconds into column 29.
        """
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(testdata['sfile_seconds_overflow'])
        event = cat[0]
        pick_times = {
            "LSb2": UTCDateTime(2009, 7, 2, 6, 49) + 100.24}
        for key, value in pick_times.items():
            pick = [p for p in event.picks
                    if p.waveform_id.station_code == key]
            assert len(pick) == 1
            assert pick[0].time == value
        with NamedTemporaryFile(suffix=".out") as tf:
            write_select(cat, filename=tf.name, nordic_format='OLD')
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                cat_back = read_events(tf.name)
        assert len(cat_back) == 1
        for key, value in pick_times.items():
            pick = [p for p in cat_back[0].picks
                    if p.waveform_id.station_code == key]
            assert len(pick) == 1
            assert pick[0].time == value

    def test_read_bad_covariance(self, testdata):
        """
        Verify graceful exit if covariance matrix is not positive-definite
        """
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(testdata['sfile_bad_covariance'])
        assert cat[0].origins[0].origin_uncertainty is None

    def test_read_high_accuracy(self, testdata):
        """
        Verify that high-accuracy locations are read, if present
        """
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(testdata['sfile_highaccuracy'])
        event = cat[0]
        event_time = event.origins[0].time
        event_lat = event.origins[0].latitude
        event_lon = event.origins[0].longitude
        event_depth = event.origins[0].depth
        event_rms = event.origins[0].quality.standard_error
        assert event_time == UTCDateTime(2015, 4, 24, 15, 25, 37) + 0.676
        assert event_lat == 37.29242
        assert event_lon == -32.26983
        assert event_depth == 1969.
        assert event_rms == 0.051

    def test_ellipse_from__to_uncerts(self):
        """
        Verify ellipse is properly calculated and inverted using uncertainties

        tests Ellipse.from_uncerts and Ellipse.to_uncerts()
        """
        center = (20, 30)
        # First try simple cases without correlation
        x_errs = (0.5, 1.33, 1.0)
        y_errs = (1.33, 0.5, 1.0)
        for c_xy in [0, 0.2, 0.4, 0.6]:
            for (x_err, y_err) in zip(x_errs, y_errs):
                ell = Ellipse.from_uncerts(x_err, y_err, c_xy, center)
                (x_err_out, y_err_out, c_xy_out, center_out) = ell.to_uncerts()
                assert round(abs(x_err-x_err_out), 7) == 0
                assert round(abs(y_err-y_err_out), 7) == 0
                assert round(abs(c_xy-c_xy_out), 7) == 0
                assert np.allclose(np.array(center), np.array(center_out))
        # Now a specific case with a finite covariance
        x_err = 0.5
        y_err = 1.1
        c_xy = -0.2149
        # Calculate ellipse
        ell = Ellipse.from_uncerts(x_err, y_err, c_xy, center)
        assert round(abs(ell.a-1.120674193646), 7) == 0
        assert round(abs(ell.b-0.451762494786), 7) == 0
        assert round(abs(ell.theta-167.9407699), 7) == 0
        # Calculate covariance error from ellipse
        (x_err_out, y_err_out, c_xy_out, center_out) = ell.to_uncerts()
        assert round(abs(x_err-x_err_out), 7) == 0
        assert round(abs(y_err-y_err_out), 7) == 0
        assert round(abs(c_xy-c_xy_out), 7) == 0
        assert np.allclose(np.array(center), np.array(center_out))

    def test_ellipse_from_to_cov(self):
        """
        Verify ellipse is properly calculated and inverted using covariance

        tests Ellipse.from_uncerts and Ellipse.to_uncerts()
        """
        center = (20, 30)
        x_err = 0.5
        y_err = 1.1
        c_xy = -0.2149
        cov = [[x_err**2, c_xy], [c_xy, y_err**2]]
        # Calculate ellipse
        ell = Ellipse.from_cov(cov, center)
        assert round(abs(ell.a-1.120674193646), 7) == 0
        assert round(abs(ell.b-0.451762494786), 7) == 0
        assert round(abs(ell.theta-167.9407699), 7) == 0
        # Calculate covariance error from ellipse
        cov_out, center_out = ell.to_cov()
        assert round(abs(cov[0][0]-cov_out[0][0]), 7) == 0
        assert round(abs(cov[0][1]-cov_out[0][1]), 7) == 0
        assert round(abs(cov[1][0]-cov_out[1][0]), 7) == 0
        assert round(abs(cov[1][1]-cov_out[1][1]), 7) == 0

    def test_ellipse_from_uncerts_baz(self, debug=False):
        """
        Verify alternative ellipse creator

        tests Ellipse.from_uncerts_baz
        """
        # Now a specific case with a finite covariance
        x_err = 0.5
        y_err = 1.1
        c_xy = -0.2149
        dist = 10
        baz = 90
        viewpoint = (5, 5)
        # Calculate ellipse
        ell = Ellipse.from_uncerts_baz(x_err, y_err, c_xy,
                                       dist, baz, viewpoint)
        assert round(abs(ell.a-1.120674193646), 7) == 0
        assert round(abs(ell.b-0.451762494786), 7) == 0
        assert round(abs(ell.theta-167.9407699), 7) == 0
        assert round(abs(ell.x-15), 7) == 0
        assert round(abs(ell.y-5), 7) == 0
        baz = 180
        ell = Ellipse.from_uncerts_baz(x_err, y_err, c_xy,
                                       dist, baz, viewpoint)
        assert round(abs(ell.x-5), 7) == 0
        assert round(abs(ell.y--5), 7) == 0

    def test_ellipse_is_inside(self, debug=False):
        """
        Verify Ellipse.is_inside()
        """
        ell = Ellipse(20, 10, 90)
        assert ell.is_inside((0, 0)) is True
        assert not ell.is_inside((100, 100))
        assert ell.is_inside((-19.9, 0))
        assert ell.is_inside((19.9, 0))
        assert not ell.is_inside((-20.1, 0))
        assert not ell.is_inside((20.1, 0))
        assert ell.is_inside((0, 9.9))
        assert ell.is_inside((0, -9.9))
        assert not ell.is_inside((0, 10.1))
        assert not ell.is_inside((0, -10.1))

    def test_ellipse_is_on(self, debug=False):
        """
        Verify Ellipse.is_on()
        """
        ell = Ellipse(20, 10, 90)
        assert not ell.is_on((0, 0))
        assert not ell.is_on((100, 100))
        assert ell.is_on((-20, 0))
        assert ell.is_on((20, 0))
        assert not ell.is_on((-20.1, 0))
        assert not ell.is_on((20.1, 0))
        assert ell.is_on((0, 10))
        assert ell.is_on((0, -10))
        assert not ell.is_on((0, 10.1))
        assert not ell.is_on((0, -10.1))

    def test_ellipse_subtended_angle(self, debug=False):
        """
        Verify Ellipse.subtended_angle()
        """
        ell = Ellipse(20, 10, 90)
        assert round(abs(ell.subtended_angle((20, 0))-180.), 7) == 0
        assert round(abs(ell.subtended_angle((0, 0))-360.), 7) == 0
        assert round(abs(ell.subtended_angle((40, 0))-32.20422750397), 7) == 0
        assert round(abs(ell.subtended_angle((0, 40))-54.62345984805), 7) == 0
        assert round(abs(ell.subtended_angle((20, 10))-89.9994270422), 7) == 0

    def test_ellipse_plot_1(self, image_path):
        """
        Test Ellipse.plot()

        To generate test figures, used same commands after:
        from ellipse import Ellipse
        import matplotlib.pyplot as plt
        plt.style.use('classic')
        """
        # Test single ellipse
        Ellipse(20, 10, 90).plot(outfile=image_path)

    def test_ellipse_plot_2(self, image_path):
        """
        Test Ellipse.plot() with multi-ellipses
        """

        # Test multi-ellipse figure
        fig = Ellipse(20, 10, 90).plot(color='r')
        fig = Ellipse(20, 10, 45).plot(fig=fig, color='b')
        fig = Ellipse(20, 10, 0, center=(10, 10)).plot(fig=fig, color='g')
        _ = Ellipse(20, 10, -45).plot(fig=fig, outfile=image_path)

    def test_ellipse_plot_tangents_1(self, image_path):
        """
        Test Ellipse.plot_tangents()
        """
        # Test single ellipse and point
        Ellipse(20, 10, 90).plot_tangents((30, 30),
                                          color='b',
                                          print_angle=True,
                                          ellipse_name='Ellipse',
                                          outfile=image_path)

    @pytest.fixture()
    def ellispe_color_cycle(self):
        """Fixture to get collors for ellipse plot."""
        import matplotlib.pyplot as plt
        prop_cycle = plt.rcParams['axes.prop_cycle']
        colors = prop_cycle.by_key()['color']
        color_cycle = cycle(colors)
        return color_cycle

    def test_ellipse_plot_tangents_2(self, image_path, ellispe_color_cycle):
        """
        Test Ellipse.plot_tangents() with multi-ellipse figure
        """
        dist = 50
        fig = None
        step = 45
        for angle in range(step, 360 + step - 1, step):
            x = dist * np.sin(np.radians(angle))
            y = dist * np.cos(np.radians(angle))
            ell = Ellipse(20, 10, 90, center=(x, y))
            if angle == 360:
                outfile = image_path
            else:
                outfile = None
            fig = ell.plot_tangents((0, 0),
                                    fig=fig,
                                    color=next(ellispe_color_cycle),
                                    print_angle=True,
                                    ellipse_name='E{:d}'.format(angle),
                                    outfile=outfile)

    def test_ellipse_plot_tangents_3(self, image_path, ellispe_color_cycle):
        """
        Test Ellipse.plot_tangents() with multi-stations.
        """
        dist = 50
        fig = None
        step = 45
        for angle in range(step, 360 + step - 1, step):
            x = dist * np.sin(np.radians(angle))
            y = dist * np.cos(np.radians(angle))
            ell = Ellipse(20, 10, 90)
            if angle == 360:
                outfile = image_path
            else:
                outfile = None
            fig = ell.plot_tangents((x, y),
                                    fig=fig,
                                    color=next(ellispe_color_cycle),
                                    print_angle=True,
                                    pt_name='pt{:d}'.format(angle),
                                    outfile=outfile)

    def test_read_uncert_ellipse(self, testdata):
        """
        Verify that confidence ellipse is properly read/calculated from nordic
        """
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            cat = read_events(testdata['sfile_highaccuracy'])
        event = cat[0]
        val = event.origins[0].origin_uncertainty
        hor_max = val['max_horizontal_uncertainty']
        hor_min = val['min_horizontal_uncertainty']
        azi_max = val['azimuth_max_horizontal_uncertainty']
        assert round(abs(hor_max-1120.674193646), 7) == 0
        assert round(abs(hor_min-451.762494786), 7) == 0
        assert round(abs(azi_max-102.0592301), 7) == 0

    def test_read_resolve_seedid(self, testdata):
        testing_path = testdata['01-0411-15L.S201309']
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            seedid = 'XX.{}.00.H{}'
            event = read_nordic(testing_path, default_seedid=seedid)[0]
        assert len(event.origins) == 2
        assert len(event.picks) == 17
        assert event.picks[0].waveform_id.network_code == 'XX'
        assert event.picks[0].waveform_id.location_code == '00'
        assert event.picks[0].waveform_id.channel_code[0] == 'H'

    def test_read_resolve_seedid_new_format(self, testdata):
        testing_path = testdata['03-0345-23L.S202101']
        # raises "UserWarning: AIN in header, currently unsupported"
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            seedid = 'XX.{}.00.H{}'
            event = read_nordic(testing_path, default_seedid=seedid)[0]
        assert len(event.origins) == 1
        assert len(event.picks) == 53
        # no changes because net code and location given in nordic file
        assert event.picks[0].waveform_id.network_code != 'XX'


def _assert_similarity(event_1, event_2, strict=False):
    """
    Raise AssertionError if testing similarity fails
    """
    similarity_output = _test_similarity(event_1, event_2, strict=strict)
    if similarity_output:
        raise AssertionError(similarity_output)


def _test_similarity(event_1, event_2, strict=False):
    """
    Check the similarity of the components of obspy events, discounting
    resource IDs, which are not maintained in nordic files.

    Raise AssertionError if test fails

    :type event_1: obspy.core.event.event.event.Event
    :param event_1: First event
    :type event_2: obspy.core.event.event.Event
    :param event_2: Comparison event
    :type verbose: bool
    :param verbose: If true and fails will output why it fails.
    """
    # What None maps to.
    pick_default_mapper = {
        "polarity": "undecidable", "evaluation_mode": "manual",
        "phase_hint": ''}
    # Check origins
    if len(event_1.origins) != len(event_2.origins):
        return False
    if strict:
        exc_orig_keys = ["resource_id", "arrivals"]
    else:
        exc_orig_keys = [
            "resource_id", "comments", "arrivals", "method_id",
            "origin_uncertainty", "depth_type", "quality",
            "creation_info", "evaluation_mode", "latitude_errors",
            "longitude_errors", "depth_errors", "time_errors"]
    for ori_1, ori_2 in zip(event_1.origins, event_2.origins):
        for key in ori_1.keys():
            if key not in exc_orig_keys:
                if ori_1[key] != ori_2[key]:
                    return ('%s is not the same as %s for key %s' %
                            (ori_1[key], ori_2[key], key))
            elif key == "arrivals":
                if len(ori_1[key]) != len(ori_2[key]):
                    return ('%i is not the same as %i for key %s' %
                            (len(ori_1[key]), len(ori_2[key]), key))
                for arr_1, arr_2 in zip(ori_1[key], ori_2[key]):
                    for arr_key in arr_1.keys():
                        if arr_key not in ["resource_id", "pick_id",
                                           "distance"]:
                            if arr_1[arr_key] != arr_2[arr_key]:
                                return ('%s does not match %s for key %s' %
                                        (arr_1[arr_key], arr_2[arr_key],
                                         arr_key))
                    if arr_1["distance"] and round(
                            arr_1["distance"]) != round(arr_2["distance"]):
                        return ('%s does not match %s for key %s' %
                                (arr_1[arr_key], arr_2[arr_key],
                                 arr_key))
    # Check picks
    if len(event_1.picks) != len(event_2.picks):
        return 'Number of picks is not equal'
    if strict:
        exc_pick_keys = ['resource_id']
    else:
        exc_pick_keys = ['resource_id', 'waveform_id']
    for pick_1, pick_2 in zip(event_1.picks, event_2.picks):
        # Assuming same ordering of picks...
        for key in pick_1.keys():
            if key not in exc_pick_keys:
                if pick_1[key] != pick_2[key]:
                    default = pick_default_mapper.get(key, None)
                    if pick_1[key] is None:
                        if pick_2[key] == default:
                            continue
                    elif pick_2[key] is None:
                        if pick_1[key] == default:
                            continue
                    return ('%s is not the same as %s for key %s' %
                            (pick_1[key], pick_2[key], key))
            elif key == "waveform_id":
                if pick_1[key].station_code != pick_2[key].station_code:
                    return 'Station codes do not match'
                if pick_1[key].channel_code[0] != pick_2[key].channel_code[0]:
                    return 'Channel codes do not match'
                if pick_1[key].channel_code[-1] !=\
                   pick_2[key].channel_code[-1]:
                    return 'Channel codes do not match'
    # Check amplitudes
    if not len(event_1.amplitudes) == len(event_2.amplitudes):
        return 'Not the same number of amplitudes'
    if strict:
        exc_amp_keys = ["resource_id", "pick_id"]
    else:
        exc_amp_keys = ["resource_id", "pick_id", "waveform_id", "snr",
                        "magnitude_hint", 'type']
    for amp_1, amp_2 in zip(event_1.amplitudes, event_2.amplitudes):
        # Assuming same ordering of amplitudes
        for key in amp_1.keys():
            if key not in exc_amp_keys:
                if not amp_1[key] == amp_2[key]:
                    return ("{0} is not the same as {1} for key "
                            "{2}\n{3}\n{4}".format(
                                amp_1[key], amp_2[key], key, amp_1, amp_2))
            elif key == "waveform_id":
                if pick_1[key].station_code != pick_2[key].station_code:
                    return 'Station codes do not match'
                if pick_1[key].channel_code[0] != pick_2[key].channel_code[0]:
                    return 'Channel codes do not match'
                if pick_1[key].channel_code[-1] !=\
                        pick_2[key].channel_code[-1]:
                    return 'Channel codes do not match'
            elif key in ["magnitude_hint", "type"]:
                # Reading back in will define both, but input event might have
                # None
                if amp_1[key] is not None:
                    if not amp_1[key] == amp_2[key]:
                        return ('%s is not the same as %s for key %s' %
                                (amp_1[key], amp_2[key], key))
    return None


def full_test_event():
    """
    Function to generate a basic, full test event
    """
    test_event = Event()
    test_event.origins.append(Origin(
        time=UTCDateTime("2012-03-26") + 1.2, latitude=45.0, longitude=25.0,
        depth=15000))
    test_event.extra = {
        'nordic_event_id': {
            'value': '20120326000001',
            'namespace':
                'https://seis.geus.net/software/seisan/node239.html'}}
    test_event.event_descriptions.append(EventDescription())
    test_event.event_descriptions[0].text = 'LE'
    test_event.creation_info = CreationInfo(agency_id='TES')
    test_event.magnitudes.append(Magnitude(
        mag=0.1, magnitude_type='ML', creation_info=CreationInfo('TES'),
        origin_id=test_event.origins[0].resource_id))
    test_event.magnitudes.append(Magnitude(
        mag=0.5, magnitude_type='Mc', creation_info=CreationInfo('TES'),
        origin_id=test_event.origins[0].resource_id))
    test_event.magnitudes.append(Magnitude(
        mag=1.3, magnitude_type='Ms', creation_info=CreationInfo('TES'),
        origin_id=test_event.origins[0].resource_id))

    # Define the test pick
    _waveform_id_1 = WaveformStreamID(station_code='FOZ', channel_code='SHZ',
                                      network_code='NZ')
    _waveform_id_2 = WaveformStreamID(station_code='WTSZ', channel_code='BH1',
                                      network_code=' ')
    # Pick to associate with amplitude - 0
    test_event.picks = [
        Pick(waveform_id=_waveform_id_1, phase_hint='IAML',
             polarity='undecidable', time=UTCDateTime("2012-03-26") + 1.68,
             evaluation_mode="manual"),
        Pick(waveform_id=_waveform_id_1, onset='impulsive', phase_hint='PN',
             polarity='positive', time=UTCDateTime("2012-03-26") + 1.68,
             evaluation_mode="manual", backazimuth=120.2,
             horizontal_slowness=9.1),
        Pick(waveform_id=_waveform_id_1, phase_hint='IAML',
             polarity='undecidable', time=UTCDateTime("2012-03-26") + 1.68,
             evaluation_mode="manual"),
        Pick(waveform_id=_waveform_id_2, onset='impulsive', phase_hint='SG',
             polarity='undecidable', time=UTCDateTime("2012-03-26") + 1.72,
             evaluation_mode="manual"),
        Pick(waveform_id=_waveform_id_2, onset='impulsive', phase_hint='PN',
             polarity='undecidable', time=UTCDateTime("2012-03-26") + 1.62,
             evaluation_mode="automatic"),
        # Missing info shouldn't be an issue
        Pick(waveform_id=_waveform_id_2, onset=None, phase_hint='PN',
             polarity=None, time=UTCDateTime("2012-03-26") + 1.92,
             evaluation_mode=None),
        # Long-phase
        Pick(waveform_id=_waveform_id_2, onset='impulsive', phase_hint='PKiKP',
             polarity=None, time=UTCDateTime("2012-03-26") + 1.92,
             evaluation_mode=None),
    ]
    # Test a generic local magnitude amplitude pick
    test_event.amplitudes = [
        Amplitude(generic_amplitude=2.0, period=0.4,
                  pick_id=test_event.picks[0].resource_id,
                  waveform_id=test_event.picks[0].waveform_id, unit='m',
                  magnitude_hint='ML', category='point', type='AML'),
        Amplitude(generic_amplitude=10,
                  pick_id=test_event.picks[1].resource_id,
                  waveform_id=test_event.picks[1].waveform_id, type='END',
                  category='duration', unit='s', magnitude_hint='Mc',
                  snr=2.3),
        Amplitude(generic_amplitude=5.0, period=0.6,
                  pick_id=test_event.picks[2].resource_id,
                  waveform_id=test_event.picks[0].waveform_id, unit='m',
                  category='point', type='AML')]
    test_event.origins[0].arrivals = [
        Arrival(time_weight=0, phase=test_event.picks[1].phase_hint,
                pick_id=test_event.picks[1].resource_id),
        Arrival(time_weight=2, phase=test_event.picks[3].phase_hint,
                pick_id=test_event.picks[3].resource_id,
                backazimuth_residual=5, time_residual=0.2, distance=15,
                azimuth=25, takeoff_angle=10),
        Arrival(time_weight=2, phase=test_event.picks[4].phase_hint,
                pick_id=test_event.picks[4].resource_id,
                backazimuth_residual=5, time_residual=0.2, distance=15,
                azimuth=25, takeoff_angle=170),
        Arrival(time_weight=2, phase=test_event.picks[5].phase_hint,
                pick_id=test_event.picks[5].resource_id,
                backazimuth_residual=5, time_residual=0.2, distance=15,
                azimuth=25, takeoff_angle=170),
        Arrival(time_weight=0, phase=test_event.picks[6].phase_hint,
                pick_id=test_event.picks[6].resource_id,
                backazimuth_residual=5, time_residual=0.2, distance=15,
                azimuth=25, takeoff_angle=170),
    ]
    # Add in error info (line E)
    test_event.origins[0].quality = OriginQuality(
        standard_error=0.01, azimuthal_gap=36)
    # Origin uncertainty in Seisan is output as long-lat-depth, quakeML has
    # semi-major and semi-minor
    test_event.origins[0].origin_uncertainty = OriginUncertainty(
        confidence_ellipsoid=ConfidenceEllipsoid(
            semi_major_axis_length=3000, semi_minor_axis_length=1000,
            semi_intermediate_axis_length=2000, major_axis_plunge=20,
            major_axis_azimuth=100, major_axis_rotation=4))
    test_event.origins[0].time_errors = QuantityError(uncertainty=0.5)
    # Add in fault-plane solution info (line F) - Note have to check program
    # used to determine which fields are filled....
    test_event.focal_mechanisms.append(FocalMechanism(
        nodal_planes=NodalPlanes(nodal_plane_1=NodalPlane(
            strike=180, dip=20, rake=30, strike_errors=QuantityError(10),
            dip_errors=QuantityError(10), rake_errors=QuantityError(20))),
        method_id=ResourceIdentifier("smi:nc.anss.org/focalMechanism/FPFIT"),
        creation_info=CreationInfo(agency_id="NC"), misfit=0.5,
        station_distribution_ratio=0.8))
    # Need to test high-precision origin and that it is preferred origin.
    # Moment tensor includes another origin
    test_event.origins.append(Origin(
        time=UTCDateTime("2012-03-26") + 1.2, latitude=45.1, longitude=25.2,
        depth=14500))
    test_event.magnitudes.append(Magnitude(
        mag=0.1, magnitude_type='MW', creation_info=CreationInfo('TES'),
        origin_id=test_event.origins[-1].resource_id))
    # Moment tensors go with focal-mechanisms
    test_event.focal_mechanisms.append(FocalMechanism(
        moment_tensor=MomentTensor(
            derived_origin_id=test_event.origins[-1].resource_id,
            moment_magnitude_id=test_event.magnitudes[-1].resource_id,
            scalar_moment=100, tensor=Tensor(
                m_rr=100, m_tt=100, m_pp=10, m_rt=1, m_rp=20, m_tp=15),
            method_id=ResourceIdentifier(
                'smi:nc.anss.org/momentTensor/BLAH'))))
    return test_event



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The InvSim test suite.
"""
import ctypes as C  # NOQA
import gzip
import io
import platform

import numpy as np

from obspy import Trace, UTCDateTime, read, read_inventory
from obspy.core.util.base import NamedTemporaryFile
from obspy.core.util.misc import SuppressOutput, ptp
from obspy.core.util.testing import traces_almost_equal
from obspy.io.sac import attach_paz
from obspy.signal.headers import clibevresp
from obspy.signal.invsim import (
    cosine_taper, estimate_magnitude, evalresp, simulate_seismometer,
    evalresp_for_frequencies)
import pytest


# Seismometers defined as in Pitsa with one zero less. The corrected
# signals are in velocity, thus must be integrated to offset and take one
# zero less than pitsa (remove 1/w in frequency domain)
PAZ_WOOD_ANDERSON = {'poles': [-6.2832 - 4.7124j,
                               -6.2832 + 4.7124j],
                     'zeros': [0.0 + 0.0j] * 1,
                     'sensitivity': 1.0,
                     'gain': 1. / 2.25}

PAZ_WWSSN_SP = {'poles': [-4.0093 - 4.0093j,
                          -4.0093 + 4.0093j,
                          -4.6077 - 6.9967j,
                          -4.6077 + 6.9967j],
                'zeros': [0.0 + 0.0j] * 2,
                'sensitivity': 1.0,
                'gain': 1. / 1.0413}

PAZ_WWSSN_LP = {'poles': [-0.4189 + 0.0j,
                          -0.4189 + 0.0j,
                          -0.0628 + 0.0j,
                          -0.0628 + 0.0j],
                'zeros': [0.0 + 0.0j] * 2,
                'sensitivity': 1.0,
                'gain': 1. / 0.0271}

PAZ_KIRNOS = {'poles': [-0.1257 - 0.2177j,
                        -0.1257 + 0.2177j,
                        -83.4473 + 0.0j,
                        -0.3285 + 0.0j],
              'zeros': [0.0 + 0.0j] * 2,
              'sensitivity': 1.0,
              'gain': 1. / 1.61}

INSTRUMENTS = {'None': None,
               'kirnos': PAZ_KIRNOS,
               'wood_anderson': PAZ_WOOD_ANDERSON,
               'wwssn_lp': PAZ_WWSSN_LP,
               'wwssn_sp': PAZ_WWSSN_SP}


class TestInvSim():
    """
    Test cases for InvSim.
    """
    def test_seis_sim_vs_pitsa1(self, testdata):
        """
        Test simulate_seismometer seismometer simulation against seismometer
        simulation of Pitsa - LE3D seismometer.
        """
        # load test file
        filename = testdata['rjob_20051006.gz']
        with gzip.open(filename) as f:
            data = np.loadtxt(f)

        # paz of test file
        samp_rate = 200.0
        paz_le3d = {'poles': [-4.21 + 4.66j,
                              -4.21 - 4.66j,
                              -2.105 + 0.0j],
                    'zeros': [0.0 + 0.0j] * 3,
                    'sensitivity': 1.0,
                    'gain': 0.4}

        for id, paz in INSTRUMENTS.items():
            # simulate instrument
            datcorr = simulate_seismometer(
                data, samp_rate, paz_remove=paz_le3d, paz_simulate=paz,
                water_level=600.0, zero_mean=False, nfft_pow2=True)
            # load pitsa file
            filename = testdata['rjob_20051006_%s.gz' % id]
            with gzip.open(filename) as f:
                data_pitsa = np.loadtxt(f)
            # calculate normalized rms
            rms = np.sqrt(np.sum((datcorr - data_pitsa) ** 2) /
                          np.sum(data_pitsa ** 2))
            assert rms < 1.1e-05

    def test_seis_sim_vs_pitsa_2(self, testdata):
        """
        Test simulate_seismometer seismometer simulation against seismometer
        simulation of Pitsa - STS-2 seismometer.
        """
        # load test file
        file = testdata['rotz_20081028.gz']
        with gzip.open(file) as f:
            data = np.loadtxt(f)

        # paz of test file
        samp_rate = 200.0
        paz_sts2 = {'poles': [-0.03736 - 0.03617j,
                              -0.03736 + 0.03617j],
                    'zeros': [0.0 + 0.0j] * 2,
                    'sensitivity': 1.0,
                    'gain': 1.5}

        for id, paz in INSTRUMENTS.items():
            # simulate instrument
            datcorr = simulate_seismometer(
                data, samp_rate, paz_remove=paz_sts2, paz_simulate=paz,
                water_level=600.0, zero_mean=False, nfft_pow2=True)
            # load pitsa file
            filename = testdata['rotz_20081028_%s.gz' % id]
            with gzip.open(filename) as f:
                data_pitsa = np.loadtxt(f)
            # calculate normalized rms
            rms = np.sqrt(np.sum((datcorr - data_pitsa) ** 2) /
                          np.sum(data_pitsa ** 2))
            assert rms < 1e-04

    def test_estimate_magnitude(self, testdata):
        """
        Tests against PITSA. Note that PITSA displays microvolt, that is
        the amplitude values must be computed back into counts (for this
        stations .596microvolt/count was used). Pitsa internally calculates
        with the sensitivity 2800 of the WA. Using this we get for the
        following for event 2009-07-19 23:03::

            RTSH PITSA 2.263 ObsPy 2.294
            RTBE PITSA 1.325 ObsPy 1.363
            RMOA PITSA 1.629 ObsPy 1.675
        """
        # the poles/zeros are the same for all three stations but the overall
        # sensitivity differs, this was probably not taken into account when
        # implementing this test (the specified 'sensitivity' is for RTSH), so
        # below we use the response for station RTSH for the test
        paz = {'poles': [-4.444 + 4.444j, -4.444 - 4.444j, -1.083 + 0j],
               'zeros': [0 + 0j, 0 + 0j, 0 + 0j],
               'gain': 1.0,
               'sensitivity': 671140000.0}
        mag_rtsh = estimate_magnitude(paz, 3.34e6, 0.065, 0.255)
        assert round(abs(mag_rtsh-2.1328727151723488), 7) == 0
        mag_rtbe = estimate_magnitude(paz, 3.61e4, 0.08, 2.197)
        assert round(abs(mag_rtbe-1.1962687721890191), 7) == 0
        mag_rnon = estimate_magnitude(paz, 6.78e4, 0.125, 1.538)
        assert round(abs(mag_rnon-1.4995311686507182), 7) == 0

        # now also test using Response object to calculate amplitude
        # (use RTSH response for all three measurements, see above comment)
        # response calculated using all stages is slightly different from the
        # PAZ + overall sensitivity used above, so we get slightly different
        # values here..
        response = read_inventory(testdata['BW_RTSH.xml'],
                                  format='STATIONXML')[0][0][0].response
        mag_rtsh = estimate_magnitude(response, 3.34e6, 0.065, 0.255)
        assert round(abs(mag_rtsh-2.1179529876187635), 7) == 0
        mag_rtbe = estimate_magnitude(response, 3.61e4, 0.08, 2.197)
        assert round(abs(mag_rtbe-1.1832677953138184), 7) == 0
        mag_rnon = estimate_magnitude(response, 6.78e4, 0.125, 1.538)
        assert round(abs(mag_rnon-1.4895395665022975), 7) == 0

    # XXX: Test for really big signal is missing, where the water level is
    # actually acting
    # def test_seisSimVsPitsa2(self):
    #    from obspy.io.mseed import tests as tests_
    #    path = os.path.dirname(__file__)
    #    file = os.path.join(path, 'data', 'BW.BGLD..EHE.D.2008.001')
    #    g = Trace()
    #    g.read(file,format='MSEED')
    #    # paz of test file
    #    samp_rate = 200.0

    def test_sac_instrument_correction(self, testdata):
        # SAC recommends to taper the transfer function if a pure
        # deconvolution is done instead of simulating a different
        # instrument. This test checks the difference between the
        # result from removing the instrument response using SAC or
        # ObsPy. Visual inspection shows that the traces are pretty
        # much identical but differences remain (rms ~ 0.042). Haven't
        # found the cause for those, yet. One possible reason is the
        # floating point arithmetic of SAC vs. the double precision
        # arithmetic of Python. However differences still seem to be
        # too big for that.
        pzf = testdata['SAC_PZs_KARC_BHZ']
        sacf = testdata['KARC.LHZ.SAC.asc.gz']
        testsacf = testdata['KARC_corrected.sac.asc.gz']
        plow = 160.
        phigh = 4.
        fl1 = 1.0 / (plow + 0.0625 * plow)
        fl2 = 1.0 / plow
        fl3 = 1.0 / phigh
        fl4 = 1.0 / (phigh - 0.25 * phigh)
        # Uncomment the following to run the sac-commands
        # that created the testing file
        # if 1:
        #    import subprocess as sp
        #    p = sp.Popen('sac',shell=True,stdin=sp.PIPE)
        #    cd1 = p.stdin
        #    print("r %s"%sacf, file=cd1)
        #    print("rmean", file=cd1)
        #    print("rtrend", file=cd1)
        #    print("taper type cosine width 0.03", file=cd1)
        #    print("transfer from polezero subtype %s to none \
        #    freqlimits %f %f %f %f" % (pzf, fl1, fl2, fl3, fl4), file=cd1)
        #    print("w over ./data/KARC_corrected.sac", file=cd1)
        #    print("quit", file=cd1)
        #    cd1.close()
        #    p.wait()

        stats = {'network': 'KA', 'delta': 0.99999988079072466,
                 'station': 'KARC', 'location': 'S1',
                 'starttime': UTCDateTime(2001, 2, 13, 0, 0, 0, 993700),
                 'calib': 1.00868e+09, 'channel': 'BHZ'}
        with gzip.open(sacf) as f:
            tr = Trace(np.loadtxt(f), stats)

        attach_paz(tr, pzf, tovel=False)
        tr.data = simulate_seismometer(
            tr.data, tr.stats.sampling_rate, paz_remove=tr.stats.paz,
            remove_sensitivity=False, pre_filt=(fl1, fl2, fl3, fl4))

        with gzip.open(testsacf) as f:
            data = np.loadtxt(f)

        # import matplotlib.pyplot as plt
        # plt.plot(tr.data)
        # plt.plot(data)
        # plt.show()
        rms = np.sqrt(np.sum((tr.data - data) ** 2) /
                      np.sum(tr.data ** 2))
        assert rms < 0.0421

    def test_evalresp_vs_obspy(self, testdata):
        """
        Compare results from removing instrument response using
        evalresp in SAC and ObsPy. Visual inspection shows that the traces are
        pretty much identical but differences remain (rms ~ 0.042). Haven't
        found the cause for those, yet.
        """
        evalrespf = testdata['CRLZ.HHZ.10.NZ.SAC_resp']
        rawf = testdata['CRLZ.HHZ.10.NZ.SAC']
        respf = testdata['RESP.NZ.CRLZ.10.HHZ']
        fl1 = 0.00588
        fl2 = 0.00625
        fl3 = 30.
        fl4 = 35.

#        #Set the following if-clause to True to run
#        #the sac-commands that created the testing file
#        if False:
#            import subprocess as sp
#            p = sp.Popen('sac', stdin=sp.PIPE)
#            cd1 = p.stdin
#            print("r %s" % rawf, file=cd1)
#            print("rmean", file=cd1)
#            print("taper type cosine width 0.05", file=cd1)
#            print("transfer from evalresp fname %s to vel freqlimits\
#            %f %f %f %f" % (respf, fl1, fl2, fl3, fl4), file=cd1)
#            print("w over %s" % evalrespf, file=cd1)
#            print("quit", file=cd1)
#            cd1.close()
#            p.wait()

        tr = read(rawf)[0]
        trtest = read(evalrespf)[0]
        date = UTCDateTime(2003, 11, 1, 0, 0, 0)
        seedresp = {'filename': respf, 'date': date, 'units': 'VEL',
                    'network': 'NZ', 'station': 'CRLZ', 'location': '10',
                    'channel': 'HHZ'}
        tr.data = simulate_seismometer(
            tr.data, tr.stats.sampling_rate, paz_remove=None,
            pre_filt=(fl1, fl2, fl3, fl4), seedresp=seedresp,
            taper_fraction=0.1, pitsasim=False, sacsim=True)
        tr.data *= 1e9
        rms = np.sqrt(np.sum((tr.data - trtest.data) ** 2) /
                      np.sum(trtest.data ** 2))
        assert rms < 0.0094
        # import matplotlib.pyplot as plt #plt.plot(tr.data-trtest.data,'b')
        # plt.plot(trtest.data,'g')
        # plt.figure()
        # plt.psd(tr.data,Fs=100.,NFFT=32768)
        # plt.psd(trtest.data,Fs=100.,NFFT=32768)
        # plt.figure()
        # plt.psd(tr.data - trtest.data, Fs=100., NFFT=32768)
        # plt.show()

    def test_cosine_taper(self, testdata):
        # SAC trace was generated with:
        # taper type cosine width 0.05
        for i in [99, 100]:
            sac_taper = testdata['ones_trace_%d_tapered.sac' % i]
            tr = read(sac_taper)[0]
            tap = cosine_taper(i, p=0.1, halfcosine=False, sactaper=True)
            np.testing.assert_array_almost_equal(tap, tr.data, decimal=6)

        # The following lines compare the cosine_taper result with
        # the result of the algorithm used by SAC in its taper routine
        # (taper.c)
        # freqs = np.fft.fftfreq(2**15,0.01)
        # fl1 = 0.00588
        # fl2 = 0.00625
        # fl3 = 30.0
        # fl4 = 35.0
        # npts = freqs.size
        # tap = cosine_taper(freqs.size, freqs=freqs, flimit=(fl1, fl2,
        #                                                     fl3, fl4))
        # tap2 = cosine_sac_taper(freqs, flimit=(fl1, fl2, fl3, fl4))
        # import matplotlib.pyplot as plt
        # plt.plot(tap,'b')
        # plt.plot(tap2,'g--')
        # plt.show()

    def test_evalresp_using_different_line_separator(self, testdata):
        """
        The evalresp needs a file with correct line separator, so '\n' for
        POSIX, '\r' for Mac OS, or '\r\n' for Windows. Here we check that
        evalresp reads all three formats.

        This test only checks the parsing capabilities of evalresp,
        the number of fft points used (nfft) can therefore be chosen
        small.
        """
        dt = UTCDateTime(2003, 11, 1, 0, 0, 0)
        nfft = 8
        # Linux
        respf = testdata['RESP.NZ.CRLZ.10.HHZ']
        evalresp(0.01, nfft, respf, dt)
        # Mac
        respf = testdata['RESP.NZ.CRLZ.10.HHZ.mac']
        evalresp(0.01, nfft, respf, dt)
        # Windows
        respf = testdata['RESP.NZ.CRLZ.10.HHZ.windows']
        evalresp(0.01, nfft, respf, dt)

    def test_evalresp_bug_395(self, testdata):
        """
        Was a bug due to inconstistent numerical range
        """
        resp = testdata['RESP.CH._.HHZ.gz']
        with NamedTemporaryFile() as fh:
            tmpfile = fh.name
            with gzip.open(resp) as f:
                fh.write(f.read())
            samprate = 120.0
            nfft = 56328
            args = [1.0 / samprate, nfft, tmpfile,
                    UTCDateTime(2012, 9, 4, 5, 12, 15, 863300)]
            kwargs = {'units': 'VEL', 'freq': True}
            _h, f = evalresp(*args, **kwargs)
            assert len(f) == nfft // 2 + 1

    def test_evalresp_specific_frequencies(self, testdata):
        """
        Test getting response for specific frequencies from evalresp
        """
        resp = testdata['RESP.CH._.HHZ.gz']
        # test some frequencies (results taken from routine
        # test_evalresp_bug_395)
        freqs = [0.0, 0.0021303792075, 0.21303792075, 0.63911376225,
                 2.1303792075, 21.303792075, 59.9978696208, 60.0]
        expected = [0j, -38033660.9731 + 14722854.5862j,
                    623756964.698 + 34705336.5587j,
                    625815840.91 + 11748438.5949j,
                    634173301.327 - 2261888.45356j,
                    689435074.739 - 216615642.231j,
                    -105.682658137 - 4360.67242023j,
                    -101.693155157 - 4172.61059939j,
                    ]
        with NamedTemporaryFile() as fh:
            tmpfile = fh.name
            with gzip.open(resp) as f:
                fh.write(f.read())
            samprate = 120.0
            t = UTCDateTime(2012, 9, 4, 5, 12, 15, 863300)
            h = evalresp_for_frequencies(
                1.0 / samprate, freqs, tmpfile, t, units='VEL')
        np.testing.assert_allclose(h, expected)

    # this test seems to fail sometimes with almost same numbers but slight
    # differences on Appveyor and we could not reproduce it on a local
    # machine.. so skip on windows.. (e.g. http://tests.obspy.org/101648/#1,
    # https://ci.appveyor.com/project/obspy/obspy/build/1.0.6561-master)
    @pytest.mark.skipif(
        platform.system() == "Windows",
        reason='unreproducible test fail encountered on Appveyor sometimes.')
    def test_evalresp_file_like_object(self, testdata):
        """
        Test evalresp with file like object
        """
        rawf = testdata['CRLZ.HHZ.10.NZ.SAC']
        respf = testdata['RESP.NZ.CRLZ.10.HHZ']

        tr1 = read(rawf)[0]
        tr2 = read(rawf)[0]

        date = UTCDateTime(2003, 11, 1, 0, 0, 0)
        seedresp = {'filename': respf, 'date': date, 'units': 'VEL',
                    'network': 'NZ', 'station': 'CRLZ', 'location': '10',
                    'channel': 'HHZ'}
        tr1.data = simulate_seismometer(
            tr1.data, tr1.stats.sampling_rate, seedresp=seedresp)

        with open(respf, 'rb') as fh:
            stringio = io.BytesIO(fh.read())
        seedresp['filename'] = stringio
        tr2.data = simulate_seismometer(tr2.data, tr2.stats.sampling_rate,
                                        seedresp=seedresp)
        assert traces_almost_equal(tr1, tr2)

    def test_segfaulting_resp_file(self, testdata):
        """
        Test case for a file that segfaults when compiled with clang and
        active optimization.

        As long as the test does not segfault it is ok.
        """
        filename = testdata["segfaulting_RESPs"] / "RESP.IE.LLRI..EHZ"
        date = UTCDateTime(2003, 11, 1, 0, 0, 0)
        # raises C-level EVRESP ERROR
        with SuppressOutput():
            with pytest.raises(ValueError):
                evalresp(t_samp=10.0, nfft=256, filename=filename, date=date,
                         station="LLRI", channel="EHZ", network="IE",
                         locid="*", units="VEL")

    def test_evalresp_seed_identifiers_work(self, testdata):
        """
        Asserts that the network, station, location and channel identifiers can
        be used to select difference responses.
        """
        kwargs = {"filename": str(testdata['RESP.OB.AAA._.BH_']),
                  "t_samp": 0.1, "nfft": 1024, "units": "VEL",
                  "date": UTCDateTime(2013, 1, 1), "network": "OP",
                  "station": "AAA", "locid": "", "freq": False, "debug": False}

        # Get the response for the first channel
        kwargs["channel"] = "BHE"
        response_1 = evalresp(**kwargs)

        # Get the second one. Should be different.
        kwargs["channel"] = "BHN"
        response_2 = evalresp(**kwargs)

        # The only thing that changed was the channel code. This should change
        # the response.
        rel_diff = ptp(np.abs(response_2 - response_1)) / \
            max(ptp(np.abs(response_1)), ptp(np.abs(response_2)))
        assert rel_diff > 1E-3

        # The RESP file only contains two channels.
        kwargs["channel"] = "BHZ"
        # suppress a C-level "no response found" warning
        with SuppressOutput():
            with pytest.raises(ValueError):
                evalresp(**kwargs)

    def test_evalresp_spline(self):
        """
        evr_spline was based on GPL plotutils, now replaced by LGPL spline
        library. Unittest for this function.
        """
        # char *evr_spline(int num_points, double *t, double *y,
        #                  double tension, double k,
        #                  double *xvals_arr, int num_xvals,
        #                  double **p_retvals_arr, int *p_num_retvals)
        clibevresp.evr_spline.argtypes = [
            C.c_int,  # num_points
            np.ctypeslib.ndpointer(dtype=np.float64, ndim=1,
                                   flags='C_CONTIGUOUS'),
            np.ctypeslib.ndpointer(dtype=np.float64, ndim=1,
                                   flags='C_CONTIGUOUS'),
            C.c_double,  # tension
            C.c_double,  # k
            np.ctypeslib.ndpointer(dtype=np.float64, ndim=1,
                                   flags='C_CONTIGUOUS'),
            C.c_int,  # num_xvals
            C.POINTER(C.POINTER(C.c_double)),
            C.POINTER(C.c_int)
        ]
        clibevresp.evr_spline.restype = C.c_char_p

        x = np.arange(1.2, 2.0, .1)
        n = len(x)
        y = np.sin(x)

        xi = x[:-1] + .05
        ni = len(xi)

        p_num_retvals = C.c_int(0)
        p_retvals_arr = C.POINTER(C.c_double)()
        res = clibevresp.evr_spline(n, x, y, 0.0, 1.0, xi, ni,
                                    C.byref(p_retvals_arr),
                                    C.byref(p_num_retvals))
        assert res is None
        assert ni == p_num_retvals.value
        yi = np.array([p_retvals_arr[i] for i in range(ni)])

        if False:  # visually verify
            import matplotlib.pyplot as plt
            plt.plot(x, y, 'bo-', 'Orig values')
            plt.plot(xi, yi, 'ro-', 'Cubic Spline interpolated values')
            plt.legend()
            plt.show()

        yi_ref = [0.94899576, 0.97572004, 0.9927136, 0.99978309, 0.99686554,
                  0.98398301, 0.96128491]
        assert np.allclose(yi, yi_ref, rtol=1e-7, atol=0)



# -*- coding: utf-8 -*-
"""
The sac.sacpz test suite.
"""
import io
import warnings

import numpy as np
import pytest

from obspy import read_inventory, Trace
from obspy.core.inventory.util import Equipment
from obspy.core.util import NamedTemporaryFile
from obspy.core.util.base import CatchAndAssertWarnings
from obspy.io.sac import attach_paz, attach_resp


class TestSACPZ:
    """
    """
    @pytest.fixture(autouse=True, scope="function")
    def setup(self, testdata):
        # these files were checked against data given by IRIS SACPZ web service
        # https://service.earthscope.org/irisws/sacpz/1/
        #                                query?net=IU&loc=*&cha=BH?&sta=ANMO
        # DIP seems to be systematically different in SACPZ output compared to
        # StationXML served by IRIS...
        self.file1 = testdata['IU_ANMO_00_BHZ.sacpz']
        self.file2 = testdata['IU_ANMO_BH.sacpz']

    @pytest.fixture(scope="class")
    def sacpz_with_no_sensors(self, testdata):
        expected = []
        with open(testdata['IU_ANMO_00_BHZ.sacpz']) as fh:
            for line in fh:
                if "INSTTYPE" in line:
                    line = "* INSTTYPE    : "
                if "CREATED" not in line:
                    expected.append(line.rstrip("\n"))

        return expected

    def test_write_sacpz_single_channel(self, root):
        """
        """
        path = root / 'core' / 'tests' / 'data' / 'IU_ANMO_00_BHZ.xml'
        inv = read_inventory(path)
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            inv.write(tempfile, format='SACPZ')
            with open(tempfile) as fh:
                got = fh.read()
        with open(self.file1) as fh:
            expected = fh.read()
        # remove CREATED line that changes
        got = [line for line in got.split("\n") if "CREATED" not in line]
        expected = [line for line in expected.split("\n")
                    if "CREATED" not in line]
        assert got == expected

    def test_write_sacpz_multiple_channels(self, root):
        """
        """
        path = root / 'core' / 'tests' / 'data' / 'IU_ANMO_BH.xml'
        inv = read_inventory(path)
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            inv.write(tempfile, format='SACPZ')
            with open(tempfile) as fh:
                got = fh.read()
        with open(self.file2) as fh:
            expected = fh.read()
        # remove CREATED line that changes
        got = [line for line in got.split("\n") if "CREATED" not in line]
        expected = [line for line in expected.split("\n")
                    if "CREATED" not in line]
        assert got == expected

    def test_write_sacpz_soh(self, datapath):
        path = (datapath.parent.parent.parent /
                'stationxml' / 'tests' / 'data' / 'only_soh.xml')
        inv = read_inventory(path)
        f = io.StringIO()
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            inv.write(f, format='SACPZ')
        # Testxml has 2 channels: 1 no paz, 2 unrecognized units.
        assert len(w) == 2
        # Assert warning messages contain correct warnings
        assert any('has no paz' in str(x.message) for x in w)
        assert any('has unrecognized input units' in str(x.message) for x in w)
        # Only 2 newlines are written.
        assert 2 == f.tell()

    @pytest.mark.parametrize("sensor", [None, Equipment(type=None)])
    def test_write_sacpz_no_sensor(self, sensor, sacpz_with_no_sensors):
        """
        Test sacpz writer when no sensor or sensor type are specified
        """
        inv = read_inventory("/path/to/IU_ANMO_00_BHZ.xml")
        inv.networks[0].stations[0].channels[0].sensor = sensor
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            inv.write(tempfile, format='SACPZ')
            with open(tempfile) as fh:
                got = [line.rstrip("\n") for line in fh
                       if "CREATED" not in line]

        assert got == sacpz_with_no_sensors

    def test_attach_paz(self):
        fvelhz = io.StringIO("""ZEROS 3
        -5.032 0.0
        POLES 6
        -0.02365 0.02365
        -0.02365 -0.02365
        -39.3011 0.
        -7.74904 0.
        -53.5979 21.7494
        -53.5979 -21.7494
        CONSTANT 2.16e18""")
        tr = Trace()
        attach_paz(tr, fvelhz, torad=True, todisp=True)
        np.testing.assert_array_almost_equal(tr.stats.paz['zeros'][0],
                                             - 31.616988, decimal=6)
        assert len(tr.stats.paz['zeros']) == 4

    def test_attach_paz_diff_order(self, testdata):
        pazfile = testdata['NZCRLZ_HHZ10.pz']
        tr = Trace()
        attach_paz(tr, pazfile)
        np.testing.assert_array_almost_equal(tr.stats.paz['gain'],
                                             7.4592e-2, decimal=6)
        assert len(tr.stats.paz['zeros']) == 5
        assert len(tr.stats.paz['poles']) == 4

    def test_sacpaz_from_dataless(self, testdata):
        # The following dictionary is extracted from a datalessSEED
        # file
        pazdict = {'sensitivity': 2516580000.0,
                   'digitizer_gain': 1677720.0, 'seismometer_gain': 1500.0,
                   'zeros': [0j, 0j], 'gain': 59198800.0,
                   'poles': [(-0.037010000000000001 + 0.037010000000000001j),
                             (-0.037010000000000001 - 0.037010000000000001j),
                             (-131 + 467.30000000000001j),
                             (-131 - 467.30000000000001j),
                             (-251.30000000000001 + 0j)]}
        tr = Trace()
        # This file was extracted from the datalessSEED file using rdseed
        pazfile = testdata['SAC_PZs_NZ_HHZ_10']
        attach_paz(tr, pazfile, todisp=False)
        sacconstant = pazdict['digitizer_gain'] * \
            pazdict['seismometer_gain'] * pazdict['gain']
        np.testing.assert_almost_equal(tr.stats.paz['gain'] / 1e17,
                                       sacconstant / 1e17, decimal=6)
        # pole-zero files according to the SAC convention are in displacement
        assert len(tr.stats.paz['zeros']) == 3

    def test_sacpaz_from_resp(self, testdata):
        # The following two files were both extracted from a dataless
        # seed file using rdseed
        respfile = testdata['RESP.NZ.CRLZ.10.HHZ']
        sacpzfile = testdata['SAC_PZs_NZ_CRLZ_HHZ']
        # This is a rather lengthy test, in which the
        # poles, zeros and the gain of each instrument response file
        # are converted into the corresponding velocity frequency response
        # function which have to be sufficiently close. Possibly due to
        # different truncations in the RESP-formatted and SAC-formatted
        # response files the frequency response functions are not identical.
        tr1 = Trace()
        tr2 = Trace()
        attach_resp(tr1, respfile, torad=True, todisp=False)
        attach_paz(tr2, sacpzfile, torad=False, tovel=True)
        p1 = tr1.stats.paz.poles
        z1 = tr1.stats.paz.zeros
        g1 = tr1.stats.paz.gain
        t_samp = 0.01
        n = 32768
        fy = 1 / (t_samp * 2.0)
        # start at zero to get zero for offset/ DC of fft
        f = np.arange(0, fy + fy / n, fy / n)  # arange should includes fy
        w = f * 2 * np.pi
        s = 1j * w
        a1 = np.poly(p1)
        b1 = g1 * np.poly(z1)
        h1 = np.polyval(b1, s) / np.polyval(a1, s)
        h1 = np.conj(h1)
        h1[-1] = h1[-1].real + 0.0j
        p2 = tr2.stats.paz.poles
        z2 = tr2.stats.paz.zeros
        g2 = tr2.stats.paz.gain
        a2 = np.poly(p2)
        b2 = g2 * np.poly(z2)
        h2 = np.polyval(b2, s) / np.polyval(a2, s)
        h2 = np.conj(h2)
        h2[-1] = h2[-1].real + 0.0j
        amp1 = abs(h1)
        amp2 = abs(h2)
        phase1 = np.unwrap(np.arctan2(-h1.imag, h1.real))
        phase2 = np.unwrap(np.arctan2(-h2.imag, h2.real))
        np.testing.assert_almost_equal(phase1, phase2, decimal=4)
        rms = np.sqrt(np.sum((amp1 - amp2) ** 2) /
                      np.sum(amp2 ** 2))
        assert rms < 2.02e-06
        assert tr1.stats.paz.t_shift == 0.4022344
        # The following plots the comparison between the
        # two frequency response functions.
        # import pylab as plt
        # plt.subplot(1,2,1)
        # plt.loglog(f,amp1)
        # plt.loglog(f,amp2,'k--')
        # plt.subplot(1,2,2)
        # plt.semilogx(f,phase1)
        # plt.semilogx(f,phase2,'k--')
        # plt.show()

    def test_writing_sacpz_hertz_to_radians(self, root):
        """
        Tests writing out a response with poles and zeros described in Hertz as
        a SACPZ file, which implicitely expects the given data to be RADIANS/S,
        so that a conversion is needed.

        See #3334
        """
        path = root / 'core' / 'tests' / 'data' / 'G_CAN__LHZ.xml'
        inv = read_inventory(path, 'STATIONXML')

        expected_a0 = 3.959488e+03
        expected_constant = 7.304622e+12
        expected_zeros = [0j, 0j, 0j]
        expected_poles = [-1.233948e-02+1.234319e-02j,
                          -1.233948e-02-1.234319e-02j,
                          -3.917566e+01+4.912339e+01j,
                          -3.917566e+01-4.912339e+01j]

        sio = io.StringIO()
        # ignore a warning because there is an additional unity PAZ stage
        with CatchAndAssertWarnings():
            inv.write(sio, format='SACPZ')
        sio.seek(0)
        lines = sio.readlines()
        # make sure we find the lines we are looking for
        for expected_start in ('* A0  ', 'ZEROS', 'POLES', 'CONSTANT'):
            for line in lines:
                if line.startswith(expected_start):
                    break
            else:
                msg = f"No line starting with '{expected_start}' found."
                pytest.fail(msg)
        # now test values
        for i, line in enumerate(lines):
            if line.startswith('* A0  '):
                value = float(line.split()[-1])
                assert round(value, 3) == expected_a0
            elif line.startswith('ZEROS'):
                num_zeros = int(line.split()[-1])
                assert num_zeros == 3
                zeros = [complex(*map(float, lines[i+1+j].split()))
                         for j in range(num_zeros)]
                np.testing.assert_allclose(zeros, expected_zeros)
            elif line.startswith('POLES'):
                num_poles = int(line.split()[-1])
                assert num_poles == 4
                poles = [complex(*map(float, lines[i+1+j].split()))
                         for j in range(num_poles)]
                np.testing.assert_allclose(poles, expected_poles)
            elif line.startswith('CONSTANT'):
                value = float(line.split()[-1])
                assert round(value, 3) == expected_constant



from obspy.taup import plot_travel_times
import matplotlib.pyplot as plt

fig, ax = plt.subplots()
ax = plot_travel_times(source_depth=10, ax=ax, fig=fig,
                       phase_list=['P', 'PP', 'S'])



# -*- coding: utf-8 -*-
"""
obspy.io.kinemetrics - Evt format support for ObsPy
===================================================

Evt read support for ObsPy.

This module provides read support for the Evt Kinemetrics data format.
It is based on the Kinemetrics description of the format and the provided
C code (Kw2asc.c (see "KW2ASC.SRC" File in /doc section)).

:copyright:
    The ObsPy Development Team (devs@obspy.org), Henri Martin, Thomas Lecocq,
    Kinemetrics(c)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

Reading
-------
Similar to reading any other waveform data format using obspy.core:

    >>> from obspy import read
    >>> st = read("/path/to/BI008_MEMA-04823.evt", apply_calib=False)
    >>> st  # doctest: +ELLIPSIS
    <obspy.core.stream.Stream object at 0x...>
    >>> print (st)  # doctest: +NORMALIZE_WHITESPACE
    3 Trace(s) in Stream:
    .MEMA..0 | 2013-08-15T09:20:28.000000Z - 2013-08-15T09:20:50.996000Z |
    250.0 Hz, 5750 samples
    .MEMA..1 | 2013-08-15T09:20:28.000000Z - 2013-08-15T09:20:50.996000Z |
    250.0 Hz, 5750 samples
    .MEMA..2 | 2013-08-15T09:20:28.000000Z - 2013-08-15T09:20:50.996000Z |
    250.0 Hz, 5750 samples

Each trace will have a ``stats`` attribute containing the usual information and
a ``kinemetrics_evt`` dictionary with specific attributes.

.. note::
    All the Header's attributes are not read (can be implemented if necessary
    for someone.)

.. code-block:: python

    >>> stats_evt = st[0].stats.pop('kinemetrics_evt')
    >>> print(st[0].stats)  # doctest: +NORMALIZE_WHITESPACE
             network:
             station: MEMA
            location:
             channel: 0
           starttime: 2013-08-15T09:20:28.000000Z
             endtime: 2013-08-15T09:20:50.996000Z
       sampling_rate: 250.0
               delta: 0.004
                npts: 5750
               calib: 1.1694431304931641e-06
             _format: KINEMETRICS_EVT
    >>> for k, v in sorted(stats_evt.items()):
    ...     print(k, v)
    ...     # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
    a2dbits 24
    acqdelay 0
    appblkversion 303
    apw 0
    aqoffset 0
    batteryvoltage -134
    bootblkversion 112
    buffersize 64
    chan_altitude 0
    chan_azimuth 0
    chan_calcoil 0.0500...
    chan_channel 0
    chan_damping 0.7070...
    chan_east 0
    chan_fullscale 2.5
    chan_gain 1
    chan_id
    chan_iirtrigfilter iira IIR bandpass, 1.2 to 20Hz @ 200sps
    chan_ltaseconds 20
    chan_natfreq 196.0
    chan_north 0
    chan_range 1g
    chan_sensitivity 2.5
    chan_sensorgain 1
    chan_sensorserialnumber 0
    chan_sensorserialnumberext 0
    chan_sensortype 32
    chan_staltaprecent 10
    chan_staltaratio 1.5
    chan_stasecondstten 0.1
    chan_triggertype threshold, default
    chan_up 0
    clocksource Internal GPS
    comment MEMBACH PARAMETERS FAC+EEP/v3.02
    crc 0
    daccount 2124
    datafmt 1
    dspblkversion 102
    duration 230
    elevation 298
    errors 0
    eventnumber 0
    filterflag 0
    filtertype 0
    flags 0
    gpsaltitude 294
    gpslastdrift1 1
    gpslastdrift2 0
    gpslastlock1 2013-08-15T09:19:20.000Z
    gpslastlock2 2013-08-15T09:16:50.000Z
    gpslastturnontime1 2013-07-19T14:22:31.000Z
    gpslastturnontime2 1980-01-01T00:00:00.000Z
    gpslastupdatetime1 2013-08-03T14:37:25.000Z
    gpslastupdatetime2 1980-01-01T00:00:00.000Z
    gpslatitude 5060
    gpslockfailcount 0
    gpslongitude 600
    gpsmaxturnontime 30
    gpssoh 0
    gpsstatus Present ON
    gpsturnoninterval 0
    gpsupdatertccount 1
    headerbytes 2040
    headerversion 130
    id KMI
    installedchan 4
    instrument New Etna
    latitude 50.609...
    localoffset 0
    longitude 6.009...
    maxchannels 12
    maxpeak -19494
    maxpeakoffset 1722
    mean -20944
    minpeak -22142
    minpeakoffset 1714
    minruntime 0
    modem_autoansweroffcmd ATS0=0
    modem_autoansweroncmd ATS0=1
    modem_calloutmsg MEMBACH CALLS SEISMOLOGY
    modem_cellontime 0
    modem_cellshare 0
    modem_cellstarttime1 -1
    modem_cellstarttime2 -1
    modem_cellstarttime3 -1
    modem_cellstarttime4 -1
    modem_cellstarttime5 -1
    modem_cellwarmuptime 0
    modem_dialingprefix ATDT
    modem_dialingsuffix
    modem_flags 63
    modem_hangupcmd ATH0
    modem_initcmd AT&FE0&C1S0=1S25=10&W
    modem_maxdialattempts 15
    modem_pausebetweencalls 200
    modem_phonenumber1 023757...
    modem_phonenumber2 027903...
    modem_phonenumber3
    modem_phonenumber4
    modem_waitforconnection 45
    nchannels 3
    nscans 5750
    postevent 15
    preevent 6
    primarystorage 0
    restartsource 0
    samplebytes 3
    samplerate 0
    secondarystorage 0
    serialnumber 4823
    sps 250
    starttime 2013-08-15T09:20:28.000Z
    starttimemsec 0
    stnid MEMA
    stream_flags functional test, FT
    sysblkversion 0
    temperature 76
    timeout 0
    triggerbitmap 6
    triggertime 2013-08-15T09:20:34.600Z
    triggertimemsec 600
    txblksize 0
    txchanmap 15
    voter_number 0
    voter_type channel
    voter_weight 1
    votestodetrigger 1
    votestotrigger 1


The actual data is stored as :class:`numpy.ndarray` in the ``data`` attribute
of each trace.

    >>> type(st[0].data)  # doctest: +ELLIPSIS
    <... 'numpy.ndarray'>
    >>> print(st[0].data)  # doctest: +ELLIPSIS
    [-20920. -20980. -20922. ... -20972. -20956. -20886.]

Writing
-------
Not implemented


"""
if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
import io
import json

import pytest

from obspy.io.json.default import Default
from obspy.io.json.core import get_dump_kwargs, _write_json
from obspy.io.quakeml.core import _read_quakeml


class TestJSON():
    """Test JSON module classes and functions"""
    @pytest.fixture(autouse=True, scope="function")
    def setup(self, root):
        qml_file = (root / 'io' / 'quakeml' / 'tests' / 'data' /
                    "qml-example-1.2-RC3.xml")
        self.c = _read_quakeml(qml_file)
        self.event = self.c.events[0]

    def verify_json(self, s):
        """Test an output is a string and is JSON"""
        assert isinstance(s, str)
        j = json.loads(s)
        assert isinstance(j, dict)

    def test_default(self):
        """Test Default function class"""
        default = Default()
        assert hasattr(default, '_catalog_attrib')
        assert hasattr(default, 'OMIT_NULLS')
        assert hasattr(default, 'TIME_FORMAT')
        assert hasattr(default, '__init__')
        assert hasattr(default, '__call__')
        s = json.dumps(self.event, default=default)
        self.verify_json(s)

    def test_get_dump_kwargs(self):
        """Test getting kwargs for json.dumps"""
        kw = get_dump_kwargs()
        assert 'default' in kw
        assert 'separators' in kw
        assert isinstance(kw['default'], Default)
        assert kw['default'].OMIT_NULLS
        assert kw['separators'] == (',', ':')
        s1 = json.dumps(self.event, **kw)
        self.verify_json(s1)
        kw = get_dump_kwargs(minify=False, no_nulls=False)
        assert 'default' in kw
        assert 'separators' not in kw
        assert isinstance(kw['default'], Default)
        assert not kw['default'].OMIT_NULLS
        s2 = json.dumps(self.event, **kw)
        self.verify_json(s2)
        # Compacted version is smaller
        assert len(s1) < len(s2)

    def test_write_json(self):
        memfile = io.StringIO()
        _write_json(self.c, memfile)
        memfile.seek(0, 0)
        # Verify json module can load
        j = json.load(memfile)
        assert isinstance(j, dict)
        # Test registered method call
        memfile = io.StringIO()
        self.c.write(memfile, format="json")
        memfile.seek(0, 0)
        # Verify json module can load
        j = json.load(memfile)
        assert isinstance(j, dict)
